# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Core API to interface with the CARLA simulator."""

import abc
import atexit
import os
import queue
import random
import signal
import time
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import carla
import gym
import numpy as np
import pygame
from absl import logging

from oatomobile.core import simulator
from oatomobile.core.registry import registry
from oatomobile.simulators.carla import defaults
from oatomobile.util import carla as cutil
from oatomobile.util import graphics as gutil

# All agents are expected to return the same action type.
CARLAAction = carla.VehicleControl  # pylint: disable=no-member


class CARLASensorTypes(simulator.SensorTypes):
  """Enumeration of types of sensors."""

  FRONT_CAMERA_RGB = 0
  BIRD_VIEW_CAMERA_RGB = 1
  LIDAR = 2
  CONTROL = 3
  LOCATION = 4
  ROTATION = 5
  VELOCITY = 6
  ACCELERATION = 7
  ORIENTATION = 8
  ANGULAR_VELOCITY = 9
  SPEED_LIMIT = 10
  IS_AT_TRAFFIC_LIGHT = 11
  TRAFFIC_LIGHT_STATE = 12
  COLLISION = 13
  LANE_INVASION = 14
  GAME_STATE = 15
  REAR_CAMERA_RGB = 16
  LEFT_CAMERA_RGB = 17
  RIGHT_CAMERA_RGB = 18
  BIRD_VIEW_CAMERA_CITYSCAPES = 19
  RED_LIGHT_INVASION = 20
  ACTORS_TRACKER = 21
  GOAL = 22
  PREDICTIONS = 23


class CameraSensor(simulator.Sensor, abc.ABC):
  """Abstract class for CARLA camera sensors."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Mapping[str, str],
      *args,
      **kwargs) -> None:
    """Constructs a camera sensor with a dedicated queue."""
    super().__init__(*args, **kwargs)
    self.config = config
    self.sensor = self._spawn_sensor(hero, self.config)  # pylint: disable=no-member
    self.queue = queue.Queue()
    self.sensor.listen(self.queue.put)

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=0.0,
        high=1.0,
        shape=(
            int(self.config["attributes"].get("image_size_y")),
            int(self.config["attributes"].get("image_size_x")),
            3,
        ),
        dtype=np.float32,
    )

  def close(self) -> None:
    """Destroys the RGB camera sensor from the CARLA server."""
    self.sensor.destroy()


class CameraRGBSensor(CameraSensor):
  """Abstract class for CARLA RGB camera sensors."""

  @staticmethod
  def _spawn_sensor(
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Optional[Mapping[str, str]] = None,
  ) -> carla.ServerSideSensor:  # pylint: disable=no-member
    """Spawns a camera on `hero`.

    Args:
      hero: The agent to attach the camera on.
      config: The attribute-value pairs for the configuration
        of the sensor.

    Returns:
      The spawned a camera sensor.
    """
    return cutil.spawn_camera(hero, config, camera_type="rgb")

  def get_observation(
      self,
      frame: int,
      timeout: float,
  ) -> np.ndarray:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.
      timeout: The interval waited before stopping search
        and raising a TimeoutError.

    Returns:
      A representation of the camera view.
    """
    try:
      while True:
        data = self.queue.get(timeout=timeout)
        # Confirms synced frames.
        if data.frame == frame:
          break
      # Processes the raw sensor data to a RGB array.
      return cutil.carla_rgb_image_to_ndarray(data)
    except queue.Empty:
      logging.debug(
          "The queue of {} sensor was empty, returns a random observation".
          format(self.uuid))
      return self.observation_space.sample()


class CameraCityScapesSensor(CameraSensor):
  """Abstract class for CARLA CityScapes camera sensors."""

  @staticmethod
  def _spawn_sensor(
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Optional[Mapping[str, str]] = None,
  ) -> carla.ServerSideSensor:  # pylint: disable=no-member
    """Spawns a camera on `hero`.

    Args:
      hero: The agent to attach the camera on.
      config: The attribute-value pairs for the configuration
        of the sensor.

    Returns:
      The spawned a camera sensor.
    """
    return cutil.spawn_camera(hero, config, camera_type="semantic_segmentation")

  def get_observation(
      self,
      frame: int,
      timeout: float,
  ) -> np.ndarray:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.
      timeout: The interval waited before stopping search
        and raising a TimeoutError.

    Returns:
      A representation of the camera view.
    """
    try:
      while True:
        data = self.queue.get(timeout=timeout)
        # Confirms synced frames.
        if data.frame == frame:
          break
      # Processes the raw sensor data to a RGB array.
      return cutil.carla_cityscapes_image_to_ndarray(data)
    except queue.Empty:
      logging.debug(
          "The queue of {} sensor was empty, returns a random observation".
          format(self.uuid))
      return self.observation_space.sample()


@registry.register_sensor(name="front_camera_rgb")
class FrontCameraRGBSensor(CameraRGBSensor):
  """Front camera view sensor."""

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "front_camera_rgb"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.FRONT_CAMERA_RGB

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "FrontCameraRGBSensor":
    """Returns the default sensor."""
    return cls(hero=hero, config=defaults.FRONT_CAMERA_RGB_SENSOR_CONFIG)


@registry.register_sensor(name="rear_camera_rgb")
class RearCameraRGBSensor(CameraRGBSensor):
  """Rear camera view sensor."""

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "rear_camera_rgb"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.REAR_CAMERA_RGB

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "RearCameraRGBSensor":
    """Returns the default sensor."""
    return cls(hero=hero, config=defaults.REAR_CAMERA_RGB_SENSOR_CONFIG)


@registry.register_sensor(name="left_camera_rgb")
class LeftCameraRGBSensor(CameraRGBSensor):
  """Left camera view sensor."""

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "left_camera_rgb"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.LEFT_CAMERA_RGB

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "LeftCameraRGBSensor":
    """Returns the default sensor."""
    return cls(hero=hero, config=defaults.LEFT_CAMERA_RGB_SENSOR_CONFIG)


@registry.register_sensor(name="right_camera_rgb")
class RightCameraRGBSensor(CameraRGBSensor):
  """Right camera view sensor."""

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "right_camera_rgb"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.RIGHT_CAMERA_RGB

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "RightCameraRGBSensor":
    """Returns the default sensor."""
    return cls(hero=hero, config=defaults.RIGHT_CAMERA_RGB_SENSOR_CONFIG)


@registry.register_sensor(name="bird_view_camera_rgb")
class BirdViewCameraRGBSensor(CameraRGBSensor):
  """Bird-view camera sensor."""

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "bird_view_camera_rgb"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.BIRD_VIEW_CAMERA_RGB

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "BirdViewCameraRGBSensor":
    """Returns the default sensor."""
    return cls(hero=hero, config=defaults.BIRD_VIEW_CAMERA_RGB_SENSOR_CONFIG)


@registry.register_sensor(name="bird_view_camera_cityscapes")
class BirdViewCameraCityScapesSensor(CameraCityScapesSensor):
  """Bird-view CityScapes camera sensor."""

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "bird_view_camera_cityscapes"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.BIRD_VIEW_CAMERA_CITYSCAPES

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "BirdViewCameraCityScapesSensor":
    """Returns the default sensor."""
    return cls(
        hero=hero,
        config=defaults.BIRD_VIEW_CAMERA_CITYSCAPES_SENSOR_CONFIG,
    )


@registry.register_sensor(name="lidar")
class LIDARSensor(simulator.Sensor):
  """LIDAR, overhead sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Mapping[str, str],
      *args,
      **kwargs) -> None:
    """Constructs an overhead LIDAR sensor with a dedicated queue."""
    super().__init__(*args, **kwargs)
    self.config = config
    self.sensor = self._spawn_sensor(hero, self.config)
    self.queue = queue.Queue()
    self.sensor.listen(self.queue.put)

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "lidar"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.LIDAR

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=0.0,
        high=1.0,
        shape=(200, 200, 2),
        dtype=np.float32,
    )

  @staticmethod
  def _spawn_sensor(
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Mapping[str, str],
  ) -> carla.ServerSideSensor:  # pylint: disable=no-member
    """Spawns LIDAR overhead sensor on `hero`.

    Args:
      hero: The agent to attach the LIDAR on.
      config: The attribute-value pairs for the configuration
        of the sensor.

    Returns:
      The spawned LIDAR overhead sensor.
    """
    return cutil.spawn_lidar(hero, config)

  def get_observation(
      self,
      frame: int,
      timeout: float,
  ) -> np.ndarray:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.
      timeout: The interval waited before stopping search
        and raising a TimeoutError.

    Returns:
      An array representation of the LIDAR point cloud.
    """
    try:
      while True:
        data = self.queue.get(timeout=timeout)
        # Confirms synced frames.
        if data.frame == frame:
          break
      # Processes the raw sensor data to a RGB array.
      return cutil.carla_lidar_measurement_to_ndarray(data)
    except queue.Empty:
      logging.debug(
          "The queue of {} sensor was empty, returns a random observation".
          format(self.uuid))
      return self.observation_space.sample()

  def close(self) -> None:
    """Destroys the LIDAR sensor from the CARLA server."""
    self.sensor.destroy()

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "LIDARSensor":
    """Returns the default sensor."""
    return cls(hero=hero, config=defaults.LIDAR_SENSOR_CONFIG)


@registry.register_sensor(name="control")
class ControlSensor(simulator.Sensor):
  """CARLA vehicle control sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a control sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "control"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.CONTROL

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-1.0,
        high=1.0,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the control of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle control.
    """
    control = self._hero.get_control()
    return cutil.carla_control_to_ndarray(control)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "ControlSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="location")
class LocationSensor(simulator.Sensor):
  """CARLA vehicle location sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a location sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "location"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.LOCATION

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the location of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle location.
    """
    location = self._hero.get_transform().location
    return cutil.carla_xyz_to_ndarray(location)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "LocationSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="rotation")
class RotationSensor(simulator.Sensor):
  """CARLA vehicle rotation sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a rotation sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "rotation"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.ROTATION

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=0.0,
        high=360.0,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the rotation of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle rotation.
    """
    rotation = self._hero.get_transform().rotation
    return cutil.carla_rotation_to_ndarray(rotation)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "RotationSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="velocity")
class VelocitySensor(simulator.Sensor):
  """CARLA vehicle velocity sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a velocity sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "velocity"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.VELOCITY

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the velocity of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle velocity.
    """
    velocity = self._hero.get_velocity()
    return cutil.carla_xyz_to_ndarray(velocity)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "VelocitySensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="acceleration")
class AccelerationSensor(simulator.Sensor):
  """CARLA vehicle acceleration sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a acceleration sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "acceleration"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.ACCELERATION

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the acceleration of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle acceleration.
    """
    acceleration = self._hero.get_acceleration()
    return cutil.carla_xyz_to_ndarray(acceleration)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "AccelerationSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="orientation")
class OrientationSensor(simulator.Sensor):
  """CARLA vehicle orientation sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a orientation sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "orientation"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.ORIENTATION

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the orientation of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle orientation.
    """
    orientation = self._hero.get_transform().get_forward_vector()
    return cutil.carla_xyz_to_ndarray(orientation)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "OrientationSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="angular_velocity")
class AngularVelocitySensor(simulator.Sensor):
  """CARLA vehicle angular velocity sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a velocity sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "angular_velocity"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.ANGULAR_VELOCITY

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the velocity of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle angular velocity.
    """
    angular_velocity = self._hero.get_angular_velocity()
    return cutil.carla_xyz_to_ndarray(angular_velocity)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "AngularVelocitySensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="speed_limit")
class SpeedLimitSensor(simulator.Sensor):
  """CARLA vehicle speed limit sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a speed limit sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "speed_limit"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.SPEED_LIMIT

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(3,),
        dtype=np.float32,
    )

  def get_observation(self, *args: Any, **kwargs: Any) -> np.ndarray:
    """Collects the speed limit of the ego vehicle.

    Returns:
      An array representation of the CARLA vehicle speed limit.
    """
    speed_limit = self._hero.get_speed_limit()
    return np.asarray(
        speed_limit,
        dtype=np.float32,
    )

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "SpeedLimitSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="is_at_traffic_light")
class IsAtTrafficLightSensor(simulator.Sensor):
  """CARLA sensor for detecting proximity from a traffic light."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a speed limit sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "is_at_traffic_light"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.IS_AT_TRAFFIC_LIGHT

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Discrete:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Discrete(n=2)

  def get_observation(self, *args: Any, **kwargs: Any) -> int:
    """Returns True if close to a traffic light.

    Returns:
      An array representation of the proximity to a traffic light.
    """
    is_at_traffic_light = self._hero.is_at_traffic_light()
    return int(is_at_traffic_light)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "IsAtTrafficLightSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="traffic_light_state")
class TrafficLightStateSensor(simulator.Sensor):
  """CARLA sensor for parsing the traffic light state."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a traffic light state sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "traffic_light_state"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.TRAFFIC_LIGHT_STATE

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Discrete:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Discrete(n=4)

  def get_observation(self, *args: Any, **kwargs: Any) -> int:
    """Returns True if close to a traffic light.

    Returns:
      An array representation of the proximity to a traffic light.
    """
    traffic_light_state = self._hero.get_traffic_light_state().conjugate()
    return int(traffic_light_state)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "TrafficLightStateSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="collision")
class CollisionSensor(simulator.Sensor):
  """Collision sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a collision sensor."""
    super().__init__(*args, **kwargs)
    self.sensor = self._spawn_sensor(hero)
    self.queue = queue.Queue()
    self.sensor.listen(self.queue.put)

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "collision"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.COLLISION

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Discrete(
        n=4)  # {0: no-collision, 1: vehicle, 2: pedestrian, 3: other}

  @staticmethod
  def _spawn_sensor(
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
  ) -> carla.ServerSideSensor:  # pylint: disable=no-member
    """Spawns collision sensor on `hero`.

    Args:
      hero: The agent to attach the collision sensor.

    Returns:
      The spawned collision sensor.
    """
    return cutil.spawn_collision(hero)

  def get_observation(self, frame: int, **kwargs) -> int:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.

    Returns:
      The collision type:
        0: no collision.
        1: collision with vehicle.
        2: collision with pedestrian.
        3: collision with other actor.
    """
    try:
      for event in self.queue.queue:
        # Confirms synced frames.
        if event.frame == frame:
          if "vehicle" in event.other_actor.type_id:
            return 1
          elif "walker" in event.other_actor.type_id:
            return 2
          else:
            return 3
      # Default return value.
      return 0
    except queue.Empty:
      logging.debug(
          "The queue of {} sensor was empty, returns a random observation".
          format(self.uuid))
      return self.observation_space.sample()

  def close(self) -> None:
    """Destroys the LIDAR sensor from the CARLA server."""
    self.sensor.destroy()

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "CollisionSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="lane_invasion")
class LaneInvasionSensor(simulator.Sensor):
  """Lane invasion sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a lane invasion sensor."""
    super().__init__(*args, **kwargs)
    self.sensor = self._spawn_sensor(hero)
    self.queue = queue.Queue()
    self.sensor.listen(self.queue.put)

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "lane_invasion"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.LANE_INVASION

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Discrete:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Discrete(n=2)  # {0: no-invasion, 1: lane-invasion}

  @staticmethod
  def _spawn_sensor(
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
  ) -> carla.ServerSideSensor:  # pylint: disable=no-member
    """Spawns lane invasion sensor on `hero`.

    Args:
      hero: The agent to attach the lane invasion sensor.

    Returns:
      The spawned lane invasion sensor.
    """
    return cutil.spawn_lane_invasion(hero)

  def get_observation(self, frame: int, **kwargs) -> int:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.

    Returns:
      The collision type:
        0: no invasion.
        1: lane invasion.
    """
    try:
      for event in self.queue.queue:
        # Confirms synced frames.
        if event.frame == frame:
          return 1
      # Default return value.
      return 0
    except queue.Empty:
      logging.debug(
          "The queue of {} sensor was empty, returns a random observation".
          format(self.uuid))
      return self.observation_space.sample()

  def close(self) -> None:
    """Destroys the lane invasion sensor from the CARLA server."""
    self.sensor.destroy()

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "LaneInvasionSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="actors_tracker")
class ActorsTrackerSensor(simulator.Sensor):
  """Lane invasion sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a lane invasion sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "actors_tracker"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.ACTORS_TRACKER

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Dict:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Dict(
        vehicles=gym.spaces.Dict(),  # id<->location
        pedestrians=gym.spaces.Dict(),  # id<->location
    )

  def get_observation(self, frame: int,
                      **kwargs) -> Mapping[str, Mapping[str, np.ndarray]]:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.

    Returns:
      The locations of the vehicles and pedestrians.
    """
    del frame  # Unused arg

    # Fetches all the vehicles and pedestrians in the map.
    world = self._hero.get_world()
    actors = world.get_actors()
    vehicles = actors.filter("vehicle.*")
    pedestrians = actors.filter("walker.pedestrian.*")

    # Output container.
    observation = dict(vehicles={}, pedestrians={})
    for actor_type, actors in [
        ("vehicles", vehicles),
        ("pedestrians", pedestrians),
    ]:
      for actor in actors:
        if actor.attributes["role_name"] is not "hero":
          observation[actor_type][actor.id] = cutil.carla_xyz_to_ndarray(
              actor.get_location())

    return observation

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "ActorsTrackerSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="goal")
class GoalSensor(simulator.Sensor):
  """Goal sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Mapping[str, str],
      destination: carla.Waypoint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a goal sensor."""
    super().__init__(*args, **kwargs)
    self.config = config
    self.destination = destination
    self._hero = hero

    # Parses hyperparameters.
    self._num_goals = self.config["num_goals"]
    self._sampling_radius = self.config["sampling_radius"]
    self._replan_every_steps = self.config["replan_every_steps"]

    # Stored goal.
    self._goal = None
    self._num_steps = 0

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "goal"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.GOAL

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(20, 3),
        dtype=np.float32,
    )

  def get_observation(self, frame: int,
                      **kwargs) -> Mapping[str, Mapping[str, np.ndarray]]:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.

    Returns:
      The ego-locations of the goal(s).
    """
    del frame  # Unused arg

    # Fetches hero measurements for the coordinate transformations.
    hero_transform = self._hero.get_transform()

    if self._goal is None or self._num_steps % self._replan_every_steps == 0:
      # References to CARLA objects.
      carla_world = self._hero.get_world()
      carla_map = carla_world.get_map()

      # Fetches start and end waypoints for the A* planner.
      origin = hero_transform.location
      start_waypoint = carla_map.get_waypoint(origin)
      end_waypoint = carla_map.get_waypoint(self.destination.location)

      # Caclulates global plan.
      waypoints, _, _ = cutil.global_plan(
          world=carla_world,
          origin=start_waypoint.transform.location,
          destination=end_waypoint.transform.location,
      )

      # Samples goals.
      goals_world = [waypoints[0]]
      for _ in range(self._num_goals - 1):
        goals_world.append(goals_world[-1].next(self._sampling_radius)[0])

      # Converts goals to `NumPy` arrays.
      self._goal = np.asarray([
          cutil.carla_xyz_to_ndarray(waypoint.transform.location)
          for waypoint in goals_world
      ])

    # Converts goals to ego coordinates.
    current_location = cutil.carla_xyz_to_ndarray(hero_transform.location)
    current_rotation = cutil.carla_rotation_to_ndarray(hero_transform.rotation)
    goals_local = cutil.world2local(
        current_location=current_location,
        current_rotation=current_rotation,
        world_locations=self._goal,
    )

    # Increments counter, bookkeping.
    self._num_steps += 1

    return goals_local.astype(np.float32)

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "GoalSensor":
    """Returns the default sensor."""
    if "destination" not in kwargs:
      raise ValueError("missing `destination` argument")
    return cls(
        hero=hero,
        config=defaults.GOAL_SENSOR_CONFIG,
        destination=kwargs.get("destination"),
    )


@registry.register_sensor(name="predictions")
class PredictionsSensor(simulator.Sensor):
  """Goal sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a goal sensor."""
    super().__init__(*args, **kwargs)
    self._hero = hero

    # Stored predictions.
    self._predictions = None

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "predictions"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.PREDICTIONS

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=-np.inf,
        high=np.inf,
        shape=(4, 2),
        dtype=np.float32,
    )

  @property
  def predictions(self) -> np.ndarray:
    """Returns predictions from PREVIOUS timestep."""
    return self._predictions

  @predictions.setter
  def predictions(self, value: np.ndarray) -> None:
    """Records the predictions for the sensor."""
    self._predictions = value

  def get_observation(self, frame: int,
                      **kwargs) -> Mapping[str, Mapping[str, np.ndarray]]:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.

    Returns:
      The ego-locations of the goal(s).
    """
    del frame  # Unused arg

    return self.predictions

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "PredictionsSensor":
    """Returns the default sensor."""
    return cls(hero=hero)


@registry.register_sensor(name="red_light_invasion")
class RedLightInvasion(simulator.Sensor):
  """Red light invasion sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> None:
    """Constructs a lane invasion sensor."""
    super().__init__(*args, **kwargs)
    self._hero = self._spawn_sensor(hero)

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "red_light_invasion"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.RED_LIGHT_INVASION

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Discrete(n=2)  # {0: no-invasion, 1: red-light-invasion}

  @staticmethod
  def _spawn_sensor(
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
  ) -> None:  # pylint: disable=no-member
    """Dummy call, to satisfy interface."""
    return hero

  def get_observation(self, frame: int, **kwargs) -> int:
    """Finds the data point that matches the current `frame` id.

    Args:
      frame: The synchronous simulation time step ID.

    Returns:
      The invasion type:
        0: no invasion.
        1: red traffi light invasion.
    """
    del frame

    # Skips calculations if not necessary.
    if not self._hero.is_at_traffic_light():
      return 0

    raise NotImplementedError

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "RedLightInvasion":
    """Returns the default sensor."""
    return cls(hero=hero, *args, **kwargs)


@registry.register_sensor(name="game_state")
class GameStateSensor(simulator.Sensor):
  """Game state sensor."""

  def __init__(
      self,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Mapping[str, str],
      *args,
      **kwargs) -> None:
    """Constructs a lane invasion sensor."""
    super().__init__(*args, **kwargs)
    self.config = config
    self.hero = hero
    # Static surfaces.
    self.road_mask, self.lane_boundaries_mask = self._spawn_sensor(
        self.hero,
        self.config,
    )

  def _get_uuid(self, *args: Any, **kwargs: Any) -> str:
    """Returns the universal unique identifier of the sensor."""
    return "game_state"

  def _get_sensor_type(self, *args: Any, **kwargs: Any) -> CARLASensorTypes:
    """Returns the the type of the sensor."""
    return CARLASensorTypes.GAME_STATE

  @property
  def observation_space(self, *args: Any, **kwargs: Any) -> gym.spaces.Box:
    """Returns the observation spec of the sensor."""
    return gym.spaces.Box(
        low=0,
        high=1,
        shape=(*self.road_mask.shape, 8),
        dtype=int,
    )

  @staticmethod
  def _spawn_sensor(
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      config: Mapping[str, str],
  ) -> Tuple[np.ndarray, np.ndarray]:  # pylint: disable=no-member
    """Spawns lane invasion sensor on `hero`.

    Args:
      hero: The ego vehicle agent.
      config: The attribute-value pairs for the configuration
        of the sensor.

    Returns:
      road_mask: The road game state.
      lane_boundaries_mask: The lane boundaries of the game state.
    """
    # Fetch CARLA world from ego vehicle.
    world = hero.get_world()

    # Fetch static surfaces.
    road_surface = gutil.get_road_surface(world, **config)
    lane_boundaries_surface = gutil.get_lane_boundaries_surface(world, **config)

    # Convert `PyGame` surfaces to binary tensors.
    road_ndarray = gutil.pygame_surface_to_ndarray(road_surface)
    road_mask = gutil.rgb_to_binary_mask(road_ndarray)
    lane_boundaries_ndarray = gutil.pygame_surface_to_ndarray(
        lane_boundaries_surface)
    lane_boundaries_mask = gutil.rgb_to_binary_mask(lane_boundaries_ndarray)

    return road_mask, lane_boundaries_mask

  def get_observation(self, **kwargs) -> np.ndarray:
    """Gets a snapshot of the game state.

    Returns:
      A multi-channel mask for the game state.
    """
    # Fetch CARLA world from ego vehicle.
    world = self.hero.get_world()

    # Fetch dynamic surfaces.
    vehicles_surface = gutil.get_vehicles_surface(world=world, **self.config)
    pedestrians_surface = gutil.get_pedestrians_surface(world=world,
                                                        **self.config)
    green_surface, yellow_surface, red_surface = gutil.get_traffic_lights_surface(
        world=world, **self.config)
    hero_surface = gutil.get_hero_surface(world=world,
                                          hero=self.hero,
                                          **self.config)

    # Convert `PyGame` surfaces to binary tensors.
    vehicles_ndarray = gutil.pygame_surface_to_ndarray(vehicles_surface)
    vehicles_mask = gutil.rgb_to_binary_mask(vehicles_ndarray)
    pedestrians_ndarray = gutil.pygame_surface_to_ndarray(pedestrians_surface)
    pedestrians_mask = gutil.rgb_to_binary_mask(pedestrians_ndarray)
    green_ndarray = gutil.pygame_surface_to_ndarray(green_surface)
    green_mask = gutil.rgb_to_binary_mask(green_ndarray)
    yellow_ndarray = gutil.pygame_surface_to_ndarray(yellow_surface)
    yellow_mask = gutil.rgb_to_binary_mask(yellow_ndarray)
    red_ndarray = gutil.pygame_surface_to_ndarray(red_surface)
    red_mask = gutil.rgb_to_binary_mask(red_ndarray)
    hero_ndarray = gutil.pygame_surface_to_ndarray(hero_surface)
    hero_mask = gutil.rgb_to_binary_mask(hero_ndarray)

    return np.c_[self.road_mask, self.lane_boundaries_mask, vehicles_mask,
                 pedestrians_mask, green_mask, yellow_mask, red_mask, hero_mask]

  def close(self) -> None:
    """Dummy call, to satisfy interface."""
    pass

  @classmethod
  def default(
      cls,
      hero: carla.ActorBlueprint,  # pylint: disable=no-member
      *args,
      **kwargs) -> "GameStateSensor":
    """Returns the default sensor."""
    return cls(hero=hero, config=defaults.GAME_STATE_CONFIG)


@registry.register_simulator(name="carla")
class CARLASimulator(simulator.Simulator):
  """A thin CARLA simulator wrapper."""

  def __init__(
      self,
      town: str,
      sensors: Sequence[str] = defaults.CARLA_SENSORS,
      spawn_point: Optional[Union[int, carla.Location]] = None,  # pylint: disable=no-member
      destination: Optional[Union[int, carla.Location]] = None,  # pylint: disable=no-member
      num_vehicles: int = 0,
      num_pedestrians: int = 0,
      fps: int = defaults.SIMULATOR_FPS,
      client_timeout: float = defaults.CARLA_CLIENT_TIMEOUT,
  ) -> None:
    """Constructs a CARLA simulator wrapper.

    Args:
      town: The `CARLA` town identifier.
      sensors: The set of sensors registered on the ego vehicle.
      spawn_point: The hero vehicle spawn point. If an int is
        provided then the index of the spawn point is used.
        If None, then randomly selects a spawn point every time
        from the available spawn points of each map.
      destination: The final destination. If an int is
        provided then the index of the spawn point is used.
        If None, then randomly selects a spawn point every time
        from the available spawn points of each map.
      num_vehicles: The number of vehicles to spawn.
      num_pedestrians: The number of pedestrians to spawn.
      fps: The frequency (in Hz) of the simulation.
      client_timeout: The time interval before stopping
        the search for the carla server.
    """
    # Configuration variables.
    self._town = town
    self._sensors = sensors
    self._fps = fps
    self._client_timeout = client_timeout
    self._num_vehicles = num_vehicles
    self._num_pedestrians = num_pedestrians

    # CARLA objects (lazy initialization).
    self._client = None
    self._world = None
    self._frame = None
    self._server = None
    self._frame0 = None
    self._dt = None
    self._vehicles = None
    self._pedestrians = None
    self._sensor_suite = None
    self._hero = None
    self._destination = destination
    self._spawn_point = spawn_point

    # Randomness controller.
    self._np_random = np.random.RandomState(None)  # pylint: disable=no-member

    # State of the game.
    self._observations = None
    self._time_elapsed = 0.0

    # Graphics setup.
    self._display = None
    self._clock = None

  @property
  def hero(self) -> carla.Vehicle:  # pylint: disable=no-member
    """Returns a reference to the ego car."""
    return self._hero

  @property
  def spawn_point(self) -> carla.Waypoint:  # pylint: disable=no-member
    """Returns a reference to the spawn point."""
    if self._world is None:
      raise ValueError("Make sure the environment is reset first.")
    return cutil.get_spawn_point(self._world, self._spawn_point)

  @property
  def destination(self) -> carla.Waypoint:  # pylint: disable=no-member
    """Returns a reference to the destination."""
    if self._world is None:
      raise ValueError("Make sure the environment is reset first.")
    return cutil.get_spawn_point(self._world, self._destination)

  @property
  def sensor_suite(self) -> simulator.SensorSuite:
    """Returns a refernce to the suite of sensors."""
    return self._sensor_suite

  @property
  def action_space(self) -> Any:
    """Returns the specification of the actions expected by the simulator."""
    return Optional[simulator.Action]

  @property
  def observation_space(self) -> Optional[gym.spaces.Dict]:
    """Returns the specification of the observations returned by the
    simulator."""
    if self.sensor_suite is None:
      return None
    else:
      return self.sensor_suite.observation_space

  def seed(self, seed: int) -> None:
    """Fixes the random number generator state."""
    random.seed(seed)
    self._np_random = np.random.RandomState(seed)  # pylint: disable=no-member

  def reset(self) -> simulator.Observations:
    """Resets the state of the simulation to the initial state.

    Returns:
      The initial observations.
    """
    # CARLA setup.
    self._client, self._world, self._frame, self._server = cutil.setup(
        town=self._town,
        fps=self._fps,
        client_timeout=self._client_timeout,
    )
    self._frame0 = int(self._frame)
    self._dt = self._world.get_settings().fixed_delta_seconds

    # Initializes hero agent.
    self._hero = cutil.spawn_hero(
        world=self._world,
        spawn_point=self.spawn_point,
        vehicle_id="vehicle.ford.mustang",
    )
    # Initializes the other vehicles.
    self._vehicles = cutil.spawn_vehicles(
        world=self._world,
        num_vehicles=self._num_vehicles,
    )
    # Initializes the pedestrians.
    self._pedestrians = cutil.spawn_pedestrians(
        world=self._world,
        num_pedestrians=self._num_pedestrians,
    )
    # Registers the sensors.
    self._sensor_suite = simulator.SensorSuite([
        registry.get_sensor(sensor).default(
            hero=self.hero,
            destination=self.destination,
        ) for sensor in self._sensors
    ])

    # HACK(filangel): due to the bug with the lifted vehicle and
    # the LocalPlanner, perform K=50 steps in the simulator.
    for _ in range(50):
      obs = self.step(action=None)

    return obs

  def step(self, action: simulator.Action) -> simulator.Observations:
    """Makes a step in the simulator, provided an action.

    Args:
      action: The hero vehicle's actions.

    Returns:
      The current set of observations from the sensors
      after the step from the `action`.
    """
    # Perform an `action` in the world.
    if action is not None:
      if not isinstance(action, CARLAAction):
        action = CARLAAction(**{k: float(v) for (k, v) in action.items()})
      self.hero.apply_control(action)

    # Advance the simulation by a time step.
    self._frame = self._world.tick()
    self._time_elapsed = np.asarray(
        (self._frame - self._frame0) * self._dt,
        dtype=np.float32,
    )
    if self._clock is not None:
      self._clock.tick()

    # Retrieves observations from registered sensors.
    self._observations = self._sensor_suite.get_observations(
        frame=self._frame,
        timeout=defaults.QUEUE_TIMEOUT,
    )

    return self._observations

  def render(self, mode: str = "human", **kwargs) -> Any:
    """Renders current state of the simulator."""
    if mode not in ("human", "rgb_array"):
      raise ValueError("Unrecognised mode value {} passed.".format(mode))

    if self._display is None or self._clock is None is None:
      # TODO(filangel): clean this up
      width = 0
      if "left_camera_rgb" in self._observations:
        width = width + 320
        height = 180
      if "front_camera_rgb" in self._observations:
        width = width + 320
        height = 180
      if "rear_camera_rgb" in self._observations:
        width = width + 320
        height = 180
      if "right_camera_rgb" in self._observations:
        width = width + 320
        height = 180
      if "lidar" in self._observations:
        width = width + 200
        height = 200
      if "bird_view_camera_rgb" in self._observations:
        width = width + 200
        height = 200
      if "bird_view_camera_cityscapes" in self._observations:
        width = width + 200
        height = 200
      self._display, self._clock, self._font = gutil.setup(
          width=width,
          height=height,
          render=mode == "human",
      )
    # It mutates the `self._display` `PyGame` object.
    gutil.make_dashboard(
        display=self._display,
        font=self._font,
        clock=self._clock,
        **{
            **self._observations,
            **kwargs
        },
    )

    if mode == "human":
      # Update window display.
      pygame.display.flip()
    elif mode == "rgb_array":
      # Converts surface to an RGB tensor.
      return gutil.pygame_surface_to_ndarray(self._display)

  def close(self) -> None:
    """Closes the simulator down and controls connection to CARLA server."""
    if self.sensor_suite is not None:
      self.sensor_suite.close()
      self._sensor_suite = None
    settings = self._world.get_settings()
    settings.synchronous_mode = False
    self._world.apply_settings(settings)
    logging.debug("Closes the CARLA server with process PID {}".format(
        self._server.pid))
    os.killpg(self._server.pid, signal.SIGKILL)
    atexit.unregister(lambda: os.killpg(self._server.pid, signal.SIGKILL))



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Registry is central source of truth in `oatomobile` (borrowed from
Habitat)."""

import collections
from typing import Optional

from absl import logging

from oatomobile.core.typing import Singleton


class Registry(metaclass=Singleton):
  """The `oatomobile` singleton registry object definition."""

  mapping = collections.defaultdict(dict)

  @classmethod
  def _register_impl(cls, _type, to_register, name, assert_type=None):

    def wrap(to_register):
      if assert_type is not None:
        assert issubclass(
            to_register,
            assert_type,
        ), "{} must be a subclass of {}".format(
            to_register,
            assert_type,
        )
      register_name = to_register.__name__ if name is None else name
      logging.debug("Registers {} at {}".format(register_name, _type))

      cls.mapping[_type][register_name] = to_register
      return to_register

    if to_register is None:
      return wrap
    else:
      return wrap(to_register)

  @classmethod
  def _get_impl(cls, _type, name):
    logging.debug("Fetches {} from {}".format(name, _type))
    return cls.mapping[_type].get(name, None)

  @classmethod
  def register_simulator(
      cls,
      to_register=None,
      name: Optional[str] = None,
  ):
    """Register a simulator to registry with key `name`."""
    from oatomobile.core.simulator import Simulator
    return cls._register_impl(
        "simulators",
        to_register,
        name,
        assert_type=Simulator,
    )

  @classmethod
  def register_sensor(
      cls,
      to_register=None,
      name: Optional[str] = None,
  ):
    """Register a sensor to registry with key `name`."""
    from oatomobile.core.simulator import Sensor
    return cls._register_impl(
        "sensors",
        to_register,
        name,
        assert_type=Sensor,
    )

  @classmethod
  def register_env(
      cls,
      to_register=None,
      name: Optional[str] = None,
  ):
    """Register a environemnt to registry with key `name`."""
    from oatomobile.core.rl import Env
    return cls._register_impl(
        "envs",
        to_register,
        name,
        assert_type=Env,
    )

  @classmethod
  def get_simulator(cls, name: str):
    """Fetches a registered simulator.

    Args:
      name: The uuid of the registered simulator.

    Returns:
      The `name`d registered simulator.
    """
    return cls._get_impl("simulators", name)

  @classmethod
  def get_sensor(cls, name: str):
    """Fetches a registered sensor.

    Args:
      name: The uuid of the registered sensor.

    Returns:
      The `name`d registered sensor.
    """
    return cls._get_impl("sensors", name)

  @classmethod
  def get_env(cls, name: str):
    """Fetches a registered environment.

    Args:
      name: The uuid of the registered environment.

    Returns:
      The `name`d registered environment.
    """
    return cls._get_impl("envs", name)


# Initializes the singleton registry for `oatomobile`.
registry = Registry()



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Handles the hosted CARLA autopilot expert demonstrations dataset."""

import glob
import os
import sys
import zipfile
from typing import Any
from typing import Callable
from typing import Generator
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Union

import carla
import matplotlib.pyplot as plt
import numpy as np
import tqdm
import wget
from absl import logging

from oatomobile.core.dataset import Dataset
from oatomobile.core.dataset import Episode
from oatomobile.util import carla as cutil
from oatomobile.util import graphics as gutil


class CARLADataset(Dataset):
  """The CARLA autopilot expert demonstrations dataset."""

  def __init__(
      self,
      id: str,
  ) -> None:
    """Constructs a CARLA dataset.

    Args:
      id: One of {"raw", "examples", "processed"}.
    """
    if id not in ("raw", "examples", "processed"):
      raise ValueError("Unrecognised CARLA dataset id {}".format(id))
    self.id = id
    super(CARLADataset, self).__init__()

  def _get_uuid(self) -> str:
    """Returns the universal unique identifier of the dataset."""
    return "CARLATown01Autopilot{}-v0".format(self.id)

  @property
  def info(self) -> Mapping[str, Any]:
    """The dataset description."""
    return dict(
        uuid=self.uuid,
        town="Town01",
        agent="carsuite_baselines.rulebased.Autopilot",
        noise=0.2,
    )

  @property
  def url(self) -> str:
    """The URL where the dataset is hosted."""
    return "https://www.cs.ox.ac.uk/people/angelos.filos/data/oatomobile/{}.zip".format(
        self.id)

  def download_and_prepare(self, output_dir: str) -> None:
    """Downloads and prepares the dataset from the host URL.

    Args:
      output_dir: The absolute path where the prepared dataset is stored.
    """
    # Creates the necessary output directory.
    os.makedirs(output_dir, exist_ok=True)

    # Temporary zip file to use.
    zfname = os.path.join(output_dir, "{}.zip".format(self.id))
    # Downloads dataset from Google Drive.
    logging.debug("Starts downloading '{}' dataset".format(self.id))
    wget.download(
        url=self.url,
        out=zfname,
    )
    # Unzips data.
    logging.debug("Unzips the data from {}".format(zfname))
    with zipfile.ZipFile(zfname) as zfile:
      zfile.extractall(output_dir)
    # Removes the zip file.
    logging.debug("Removes the compressed {}".format(zfname))
    os.remove(zfname)

  @staticmethod
  def load_datum(
      fname: str,
      modalities: Sequence[str],
      mode: bool,
      dataformat: str = "HWC",
  ) -> Mapping[str, np.ndarray]:
    """Loads a single datum from the dataset.

    Args:
      fname: The absolute path to the ".npz" datum.
      modalities: The keys of the attributes to fetch.
      mode: If True, it labels its datum with {FORWARD, STOP, LEFT, RIGHT}.
      dataformat: The format of the 3D data, one of `{HWC, CHW}`.

    Returns:
      The datum in a dictionary, `NumPy`-friendly format.
    """
    assert dataformat in ("HWC", "CHW")

    dtype = np.float32
    sample = dict()

    with np.load(fname) as datum:
      for attr in modalities:
        # Fetches the value.
        sample[attr] = datum[attr]
        # Converts scalars to 1D vectors.
        sample[attr] = np.atleast_1d(sample[attr])
        # Casts value to same type.
        sample[attr] = sample[attr].astype(dtype)
        if len(sample[attr].shape) == 3 and dataformat == "CHW":
          # Converts from HWC to CHW format.
          sample[attr] = np.transpose(sample[attr], (2, 0, 1))

    # Appends `mode` attribute where `{0: FORWARD, 1: STOP, 2: TURN}`.
    if mode and "player_future" in sample:
      plan = sample["player_future"]
      x_T, y_T = plan[-1, :2]
      # Norm of the vector (x_T, y_T).
      norm = np.linalg.norm([x_T, y_T])
      # Angle of vector (0, 0) -> (x_T, y_T).
      theta = np.degrees(np.arccos(x_T / (norm + 1e-3)))
      if norm < 3:  # STOP
        sample["mode"] = 1
      elif theta > 15:  # LEFT
        sample["mode"] = 2
      elif theta <= -15:  # RIGHT
        sample["mode"] = 3
      else:  # FORWARD
        sample["mode"] = 0
      sample["mode"] = np.atleast_1d(sample["mode"])
      sample["mode"] = sample["mode"].astype(dtype)

    # Records the path to the sample.
    sample["name"] = fname

    return sample

  @staticmethod
  def collect(
      town: str,
      output_dir: str,
      num_vehicles: int,
      num_pedestrians: int,
      num_steps: int = 1000,
      spawn_point: Optional[Union[int, carla.Location]] = None,  # pylint: disable=no-member
      destination: Optional[Union[int, carla.Location]] = None,  # pylint: disable=no-member
      sensors: Sequence[str] = (
          "acceleration",
          "velocity",
          "lidar",
          "is_at_traffic_light",
          "traffic_light_state",
          "actors_tracker",
      ),
      render: bool = False,
  ) -> None:
    """Collects autopilot demonstrations for a single episode on CARLA.

    Args:
      town: The CARLA town id.
      output_dir: The full path to the output directory.
      num_vehicles: The number of other vehicles in the simulation.
      num_pedestrians: The number of pedestrians in the simulation.
      num_steps: The number of steps in the simulator.
      spawn_point: The hero vehicle spawn point. If an int is
        provided then the index of the spawn point is used.
        If None, then randomly selects a spawn point every time
        from the available spawn points of each map.
      destination: The final destination. If an int is
        provided then the index of the spawn point is used.
        If None, then randomly selects a spawn point every time
        from the available spawn points of each map.
      sensors: The list of recorded sensors.
      render: If True it spawn the `PyGame` display.
    """
    from oatomobile.baselines.rulebased.autopilot.agent import AutopilotAgent
    from oatomobile.core.loop import EnvironmentLoop
    from oatomobile.core.rl import FiniteHorizonWrapper
    from oatomobile.core.rl import SaveToDiskWrapper
    from oatomobile.envs.carla import CARLAEnv
    from oatomobile.envs.carla import TerminateOnCollisionWrapper

    # Storage area.
    os.makedirs(output_dir, exist_ok=True)

    # Initializes a CARLA environment.
    env = CARLAEnv(
        town=town,
        sensors=sensors,
        spawn_point=spawn_point,
        destination=destination,
        num_vehicles=num_vehicles,
        num_pedestrians=num_pedestrians,
    )
    # Terminates episode if a collision occurs.
    env = TerminateOnCollisionWrapper(env)
    # Wraps the environment in an episode handler to store <observation, action> pairs.
    env = SaveToDiskWrapper(env=env, output_dir=output_dir)
    # Caps environment's duration.
    env = FiniteHorizonWrapper(env=env, max_episode_steps=num_steps)

    # Run a full episode.
    EnvironmentLoop(
        agent_fn=AutopilotAgent,
        environment=env,
        render_mode="human" if render else "none",
    ).run()

  @staticmethod
  def process(
      dataset_dir: str,
      output_dir: str,
      future_length: int = 80,
      past_length: int = 20,
      num_frame_skips: int = 5,
  ) -> None:
    """Converts a raw dataset to demonstrations for imitation learning.

    Args:
      dataset_dir: The full path to the raw dataset.
      output_dir: The full path to the output directory.
      future_length: The length of the future trajectory.
      past_length: The length of the past trajectory.
      num_frame_skips: The number of frames to skip.
    """

    # Creates the necessary output directory.
    os.makedirs(output_dir, exist_ok=True)

    # Iterate over all episodes.
    for episode_token in tqdm.tqdm(os.listdir(dataset_dir)):
      logging.debug("Processes {} episode".format(episode_token))
      # Initializes episode handler.
      episode = Episode(parent_dir=dataset_dir, token=episode_token)
      # Fetches all `.npz` files from the raw dataset.
      try:
        sequence = episode.fetch()
      except FileNotFoundError:
        continue

      # Always keep `past_length+future_length+1` files open.
      assert len(sequence) >= past_length + future_length + 1
      for i in tqdm.trange(
          past_length,
          len(sequence) - future_length,
          num_frame_skips,
      ):
        try:
          # Player context/observation.
          observation = episode.read_sample(sample_token=sequence[i])
          current_location = observation["location"]
          current_rotation = observation["rotation"]

          # Build past trajectory.
          player_past = list()
          for j in range(past_length, 0, -1):
            past_location = episode.read_sample(
                sample_token=sequence[i - j],
                attr="location",
            )
            player_past.append(past_location)
          player_past = np.asarray(player_past)
          assert len(player_past.shape) == 2
          player_past = cutil.world2local(
              current_location=current_location,
              current_rotation=current_rotation,
              world_locations=player_past,
          )

          # Build future trajectory.
          player_future = list()
          for j in range(1, future_length + 1):
            future_location = episode.read_sample(
                sample_token=sequence[i + j],
                attr="location",
            )
            player_future.append(future_location)
          player_future = np.asarray(player_future)
          assert len(player_future.shape) == 2
          player_future = cutil.world2local(
              current_location=current_location,
              current_rotation=current_rotation,
              world_locations=player_future,
          )

          # Store to ouput directory.
          np.savez_compressed(
              os.path.join(output_dir, "{}.npz".format(sequence[i])),
              **observation,
              player_future=player_future,
              player_past=player_past,
          )

        except Exception as e:
          if isinstance(e, KeyboardInterrupt):
            sys.exit(0)

  @staticmethod
  def plot_datum(
      fname: str,
      output_dir: str,
  ) -> None:
    """Visualizes a datum from the dataset.

    Args:
      fname: The absolute path to the datum.
      output_dir: The full path to the output directory.
    """
    COLORS = [
        "#0071bc",
        "#d85218",
        "#ecb01f",
        "#7d2e8d",
        "#76ab2f",
        "#4cbded",
        "#a1132e",
    ]

    # Creates the necessary output directory.
    os.makedirs(output_dir, exist_ok=True)

    # Load datum.
    datum = np.load(fname)

    # Draws LIDAR.
    if "lidar" in datum:
      bev_meters = 25.0
      lidar = gutil.lidar_2darray_to_rgb(datum["lidar"])
      fig, ax = plt.subplots(figsize=(3.0, 3.0))
      ax.imshow(
          np.transpose(lidar, (1, 0, 2)),
          extent=(-bev_meters, bev_meters, bev_meters, -bev_meters),
      )
      ax.set(frame_on=False)
      ax.get_xaxis().set_visible(False)
      ax.get_yaxis().set_visible(False)
      fig.savefig(
          os.path.join(output_dir, "lidar.png"),
          bbox_inches="tight",
          pad_inches=0,
          transparent=True,
      )

    # Draws first person camera-view.
    if "front_camera_rgb" in datum:
      front_camera_rgb = datum["front_camera_rgb"]
      fig, ax = plt.subplots(figsize=(3.0, 3.0))
      ax.imshow(front_camera_rgb)
      ax.set(frame_on=False)
      ax.get_xaxis().set_visible(False)
      ax.get_yaxis().set_visible(False)
      fig.savefig(
          os.path.join(output_dir, "front_camera_rgb.png"),
          bbox_inches="tight",
          pad_inches=0,
          transparent=True,
      )

    # Draws bird-view camera.
    if "bird_view_camera_cityscapes" in datum:
      bev_meters = 25.0
      bird_view_camera_cityscapes = datum["bird_view_camera_cityscapes"]
      fig, ax = plt.subplots(figsize=(3.0, 3.0))
      ax.imshow(
          bird_view_camera_cityscapes,
          extent=(-bev_meters, bev_meters, bev_meters, -bev_meters),
      )
      # Draw past if available.
      if "player_past" in datum:
        player_past = datum["player_past"]
        ax.plot(
            player_past[..., 1],
            -player_past[..., 0],
            marker="x",
            markersize=4,
            color=COLORS[0],
            alpha=0.15,
        )
      # Draws future if available.
      if "player_future" in datum:
        player_future = datum["player_future"]
        ax.plot(
            player_future[..., 1],
            -player_future[..., 0],
            marker="o",
            markersize=4,
            color=COLORS[1],
            alpha=0.15,
        )
      # Draws goals if available.
      if "goal" in datum:
        goal = datum["goal"]
        ax.plot(
            goal[..., 1],
            -goal[..., 0],
            marker="D",
            markersize=6,
            color=COLORS[2],
            linestyle="None",
            alpha=0.25,
            label=r"$\mathcal{G}$",
        )
      ax.set(frame_on=False)
      ax.get_xaxis().set_visible(False)
      ax.get_yaxis().set_visible(False)
      fig.savefig(
          os.path.join(output_dir, "bird_view_camera_cityscapes.png"),
          bbox_inches="tight",
          pad_inches=0,
          transparent=True,
      )

    # Draws bird-view camera.
    if "bird_view_camera_rgb" in datum:
      bev_meters = 25.0
      bird_view_camera_rgb = datum["bird_view_camera_rgb"]
      fig, ax = plt.subplots(figsize=(3.0, 3.0))
      ax.imshow(
          bird_view_camera_rgb,
          extent=(-bev_meters, bev_meters, bev_meters, -bev_meters),
      )
      # Draw past if available.
      if "player_past" in datum:
        player_past = datum["player_past"]
        ax.plot(
            player_past[..., 1],
            -player_past[..., 0],
            marker="x",
            markersize=4,
            color=COLORS[0],
            alpha=0.15,
        )
      # Draws future if available.
      if "player_future" in datum:
        player_future = datum["player_future"]
        ax.plot(
            player_future[..., 1],
            -player_future[..., 0],
            marker="o",
            markersize=4,
            color=COLORS[1],
            alpha=0.15,
        )
      ax.set(frame_on=False)
      ax.get_xaxis().set_visible(False)
      ax.get_yaxis().set_visible(False)
      fig.savefig(
          os.path.join(output_dir, "bird_view_camera_rgb.png"),
          bbox_inches="tight",
          pad_inches=0,
          transparent=True,
      )

  @classmethod
  def plot_coverage(
      cls,
      dataset_dir: str,
      output_fname: str,
      color: int = 0,
  ) -> None:
    """Visualizes all the trajectories in the dataset.

    Args:
      dataset_dir: The parent directory of all the dataset.
      output_fname: The full path to the output filename.
      color: The index of the color to use for the trajectories.
    """
    COLORS = [
        "#0071bc",
        "#d85218",
        "#ecb01f",
        "#7d2e8d",
        "#76ab2f",
        "#4cbded",
        "#a1132e",
    ]

    # Fetches all the data points.
    data_files = glob.glob(
        os.path.join(dataset_dir, "**", "*.npz"),
        recursive=True,
    )

    # Container that stores all locaitons.
    locations = list()
    for npz_fname in tqdm.tqdm(data_files):
      try:
        locations.append(
            cls.load_datum(
                npz_fname,
                modalities=["location"],
                mode=False,
            )["location"])
      except Exception as e:
        if isinstance(e, KeyboardInterrupt):
          sys.exit(0)
    locations = np.asarray(locations)

    # Scatter plots all locaitons.
    fig, ax = plt.subplots(figsize=(3.0, 3.0))
    ax.scatter(
        locations[..., 0],
        locations[..., 1],
        s=5,
        alpha=0.01,
        color=COLORS[color % len(COLORS)],
    )
    ax.set(title=dataset_dir, frame_on=False)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    fig.savefig(
        os.path.join(output_fname),
        bbox_inches="tight",
        pad_inches=0,
        transparent=True,
    )

  @classmethod
  def as_tensorflow(
      cls,
      dataset_dir: str,
      modalities: Sequence[str],
      mode: bool = False,
  ) -> "tensorflow.data.Dataset":
    """Implements a data reader and loader for the expert demonstrations.

    Args:
      dataset_dir: The absolute path to the raw dataset.
      modalities: The keys of the attributes to fetch.
      mode: If True, it labels its datum with {FORWARD, STOP, LEFT, RIGHT}.

    Returns:
      The unbatched `TensorFlow` dataset.
    """
    import tensorflow as tf

    # Fetches all the filenames.
    filenames = glob.glob(os.path.join(dataset_dir, "*.npz"))

    # Gets shapes of output tensors.
    output_shapes = dict()
    with np.load(filenames[0]) as datum:
      for modality in modalities:
        output_shapes[modality] = tf.TensorShape(
            np.atleast_1d(datum[modality]).shape)

    # Appends "mode" attribute.
    if mode:
      output_shapes["mode"] = tf.TensorShape((1,))

    # Sets all output types to `tf.float32`.
    output_types = {modality: tf.float32 for modality in output_shapes.keys()}

    return tf.data.Dataset.from_generator(
        generator=lambda: (cls.load_datum(
            npz_fname,
            modalities,
            mode,
            dataformat="HWC",
        ) for npz_fname in filenames),
        output_types=output_types,
        output_shapes=output_shapes,
    )

  @classmethod
  def as_numpy(
      cls,
      dataset_dir: str,
      modalities: Sequence[str],
      mode: bool = False,
  ) -> Generator[Mapping[str, np.ndarray], None, None]:
    """Implements a data reader and loader for the expert demonstrations.

    Args:
      dataset_dir: The absolute path to the raw dataset.
      modalities: The keys of the attributes to fetch.
      mode: If True, it labels its datum with {FORWARD, STOP, LEFT, RIGHT}.

    Returns:
      The unbatched `NumPy` dataset.
    """
    import tensorflow_datasets as tfds

    return tfds.as_numpy(cls.as_tensorflow(dataset_dir, modalities, mode))

  @classmethod
  def as_torch(
      cls,
      dataset_dir: str,
      modalities: Sequence[str],
      transform: Optional[Callable[[Any], Any]] = None,
      mode: bool = False,
      only_array: bool = False,
  ) -> "torch.utils.data.Dataset":
    """Implements a data reader and loader for the expert demonstrations.

    Args:
      dataset_dir: The absolute path to the raw dataset.
      modalities: The keys of the attributes to fetch.
      transform: The transformations applied on each datum.
      mode: If True, it labels its datum with {FORWARD, STOP, LEFT, RIGHT}.
      only_array: If True, it removes all the keys that are non-array, useful
        when training a model and want to run `.to(device)` without errors.

    Returns:
      The unbatched `PyTorch` dataset.
    """
    import torch

    class PyTorchDataset(torch.utils.data.Dataset):
      """Implementa a data reader for the expert demonstrations."""

      def __init__(
          self,
          dataset_dir: str,
          modalities: Sequence[str],
          transform: Optional[Callable[[Any], Any]] = None,
          mode: bool = False,
      ) -> None:
        """A simple `PyTorch` dataset.

        Args:
          dataset_dir: The absolute path to the raw dataset.
          modalities: The keys of the attributes to fetch.
          mode: If True, it labels its datum with {FORWARD, STOP, LEFT, RIGHT}.
        """
        # Internalise hyperparameters.
        self._modalities = modalities
        self._npz_files = glob.glob(os.path.join(dataset_dir, "*.npz"))
        self._transform = transform
        self._mode = mode

      def __len__(self) -> int:
        """Returns the size of the dataset."""
        return len(self._npz_files)

      def __getitem__(
          self,
          idx: int,
      ) -> Mapping[str, np.ndarray]:
        """Loads a single datum.

        Returns:
          The datum in `NumPy`-friendly format.
        """
        # Loads datum from dataset.
        sample = cls.load_datum(
            fname=self._npz_files[idx],
            modalities=self._modalities,
            mode=self._mode,
            dataformat="CHW",
        )

        # Filters out non-array keys.
        for key in list(sample):
          if not isinstance(sample[key], np.ndarray):
            sample.pop(key)

        # Applies (optional) transformation to all values.
        if self._transform is not None:
          sample = {key: self._transform(val) for (key, val) in sample.items()}
        return sample

    return PyTorchDataset(dataset_dir, modalities, transform, mode)



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Trains the deep imitative model on expert demostrations."""

import os
from typing import Mapping

import torch
import torch.distributions as D
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import tqdm
from absl import app
from absl import flags
from absl import logging

from oatomobile.baselines.torch.dim.model import ImitativeModel
from oatomobile.baselines.torch.logging import Checkpointer
from oatomobile.baselines.torch.logging import TensorBoardWriter
from oatomobile.baselines.torch.typing import ArrayLike
from oatomobile.datasets.carla import CARLADataset

logging.set_verbosity(logging.DEBUG)
FLAGS = flags.FLAGS
flags.DEFINE_string(
    name="dataset_dir",
    default=None,
    help="The full path to the processed dataset.",
)
flags.DEFINE_string(
    name="output_dir",
    default=None,
    help="The full path to the output directory (for logs, ckpts).",
)
flags.DEFINE_integer(
    name="batch_size",
    default=512,
    help="The batch size used for training the neural network.",
)
flags.DEFINE_integer(
    name="num_epochs",
    default=None,
    help="The number of training epochs for the neural network.",
)
flags.DEFINE_integer(
    name="save_model_frequency",
    default=4,
    help="The number epochs between saves of the model.",
)
flags.DEFINE_float(
    name="learning_rate",
    default=1e-3,
    help="The ADAM learning rate.",
)
flags.DEFINE_integer(
    name="num_timesteps_to_keep",
    default=4,
    help="The numbers of time-steps to keep from the target, with downsampling.",
)
flags.DEFINE_float(
    name="weight_decay",
    default=0.0,
    help="The L2 penalty (regularization) coefficient.",
)
flags.DEFINE_bool(
    name="clip_gradients",
    default=False,
    help="If True it clips the gradients norm to 1.0.",
)


def main(argv):
  # Debugging purposes.
  logging.debug(argv)
  logging.debug(FLAGS)

  # Parses command line arguments.
  dataset_dir = FLAGS.dataset_dir
  output_dir = FLAGS.output_dir
  batch_size = FLAGS.batch_size
  num_epochs = FLAGS.num_epochs
  learning_rate = FLAGS.learning_rate
  save_model_frequency = FLAGS.save_model_frequency
  num_timesteps_to_keep = FLAGS.num_timesteps_to_keep
  weight_decay = FLAGS.weight_decay
  clip_gradients = FLAGS.clip_gradients
  noise_level = 1e-2

  # Determines device, accelerator.
  device = torch.device("cuda" if torch.cuda.is_available() else "cpu")  # pylint: disable=no-member

  # Creates the necessary output directory.
  os.makedirs(output_dir, exist_ok=True)
  log_dir = os.path.join(output_dir, "logs")
  os.makedirs(log_dir, exist_ok=True)
  ckpt_dir = os.path.join(output_dir, "ckpts")
  os.makedirs(ckpt_dir, exist_ok=True)

  # Initializes the model and its optimizer.
  output_shape = [num_timesteps_to_keep, 2]
  model = ImitativeModel(output_shape=output_shape).to(device)
  optimizer = optim.Adam(
      model.parameters(),
      lr=learning_rate,
      weight_decay=weight_decay,
  )
  writer = TensorBoardWriter(log_dir=log_dir)
  checkpointer = Checkpointer(model=model, ckpt_dir=ckpt_dir)

  def transform(batch: Mapping[str, ArrayLike]) -> Mapping[str, torch.Tensor]:
    """Preprocesses a batch for the model.

    Args:
      batch: (keyword arguments) The raw batch variables.

    Returns:
      The processed batch.
    """
    # Sends tensors to `device`.
    batch = {key: tensor.to(device) for (key, tensor) in batch.items()}
    # Preprocesses batch for the model.
    batch = model.transform(batch)
    return batch

  # Setups the dataset and the dataloader.
  modalities = (
      "lidar",
      "is_at_traffic_light",
      "traffic_light_state",
      "player_future",
      "velocity",
  )
  dataset_train = CARLADataset.as_torch(
      dataset_dir=os.path.join(dataset_dir, "train"),
      modalities=modalities,
  )
  dataloader_train = torch.utils.data.DataLoader(
      dataset_train,
      batch_size=batch_size,
      shuffle=True,
      num_workers=50,
  )
  dataset_val = CARLADataset.as_torch(
      dataset_dir=os.path.join(dataset_dir, "val"),
      modalities=modalities,
  )
  dataloader_val = torch.utils.data.DataLoader(
      dataset_val,
      batch_size=batch_size * 5,
      shuffle=True,
      num_workers=50,
  )

  # Theoretical limit of NLL.
  nll_limit = -torch.sum(  # pylint: disable=no-member
      D.MultivariateNormal(
          loc=torch.zeros(output_shape[-2] * output_shape[-1]),  # pylint: disable=no-member
          scale_tril=torch.eye(output_shape[-2] * output_shape[-1]) *  # pylint: disable=no-member
          noise_level,  # pylint: disable=no-member
      ).log_prob(torch.zeros(output_shape[-2] * output_shape[-1])))  # pylint: disable=no-member

  def train_step(
      model: ImitativeModel,
      optimizer: optim.Optimizer,
      batch: Mapping[str, torch.Tensor],
      clip: bool = False,
  ) -> torch.Tensor:
    """Performs a single gradient-descent optimisation step."""
    # Resets optimizer's gradients.
    optimizer.zero_grad()

    # Perturb target.
    y = torch.normal(  # pylint: disable=no-member
        mean=batch["player_future"][..., :2],
        std=torch.ones_like(batch["player_future"][..., :2]) * noise_level,  # pylint: disable=no-member
    )

    # Forward pass from the model.
    z = model._params(
        velocity=batch["velocity"],
        visual_features=batch["visual_features"],
        is_at_traffic_light=batch["is_at_traffic_light"],
        traffic_light_state=batch["traffic_light_state"],
    )
    _, log_prob, logabsdet = model._inverse(y=y, z=z)

    # Calculates loss (NLL).
    loss = -torch.mean(log_prob - logabsdet, dim=0)  # pylint: disable=no-member

    # Backward pass.
    loss.backward()

    # Clips gradients norm.
    if clip:
      torch.nn.utils.clip_grad_norm(model.parameters(), 1.0)

    # Performs a gradient descent step.
    optimizer.step()

    return loss

  def train_epoch(
      model: ImitativeModel,
      optimizer: optim.Optimizer,
      dataloader: torch.utils.data.DataLoader,
  ) -> torch.Tensor:
    """Performs an epoch of gradient descent optimization on `dataloader`."""
    model.train()
    loss = 0.0
    with tqdm.tqdm(dataloader) as pbar:
      for batch in pbar:
        # Prepares the batch.
        batch = transform(batch)
        # Performs a gradien-descent step.
        loss += train_step(model, optimizer, batch, clip=clip_gradients)
    return loss / len(dataloader)

  def evaluate_step(
      model: ImitativeModel,
      batch: Mapping[str, torch.Tensor],
  ) -> torch.Tensor:
    """Evaluates `model` on a `batch`."""
    # Forward pass from the model.
    z = model._params(
        velocity=batch["velocity"],
        visual_features=batch["visual_features"],
        is_at_traffic_light=batch["is_at_traffic_light"],
        traffic_light_state=batch["traffic_light_state"],
    )
    _, log_prob, logabsdet = model._inverse(
        y=batch["player_future"][..., :2],
        z=z,
    )

    # Calculates loss (NLL).
    loss = -torch.mean(log_prob - logabsdet, dim=0)  # pylint: disable=no-member

    return loss

  def evaluate_epoch(
      model: ImitativeModel,
      dataloader: torch.utils.data.DataLoader,
  ) -> torch.Tensor:
    """Performs an evaluation of the `model` on the `dataloader."""
    model.eval()
    loss = 0.0
    with tqdm.tqdm(dataloader) as pbar:
      for batch in pbar:
        # Prepares the batch.
        batch = transform(batch)
        # Accumulates loss in dataset.
        with torch.no_grad():
          loss += evaluate_step(model, batch)
    return loss / len(dataloader)

  def write(
      model: ImitativeModel,
      dataloader: torch.utils.data.DataLoader,
      writer: TensorBoardWriter,
      split: str,
      loss: torch.Tensor,
      epoch: int,
  ) -> None:
    """Visualises model performance on `TensorBoard`."""
    # Gets a sample from the dataset.
    batch = next(iter(dataloader))
    # Prepares the batch.
    batch = transform(batch)
    # Turns off gradients for model parameters.
    for params in model.parameters():
      params.requires_grad = False
    # Generates predictions.
    predictions = model(num_steps=20, **batch)
    # Turns on gradients for model parameters.
    for params in model.parameters():
      params.requires_grad = True
    # Logs on `TensorBoard`.
    writer.log(
        split=split,
        loss=loss.detach().cpu().numpy().item(),
        overhead_features=batch["visual_features"].detach().cpu().numpy()[:8],
        predictions=predictions.detach().cpu().numpy()[:8],
        ground_truth=batch["player_future"].detach().cpu().numpy()[:8],
        global_step=epoch,
    )

  with tqdm.tqdm(range(num_epochs)) as pbar_epoch:
    for epoch in pbar_epoch:
      # Trains model on whole training dataset, and writes on `TensorBoard`.
      loss_train = train_epoch(model, optimizer, dataloader_train)
      write(model, dataloader_train, writer, "train", loss_train, epoch)

      # Evaluates model on whole validation dataset, and writes on `TensorBoard`.
      loss_val = evaluate_epoch(model, dataloader_val)
      write(model, dataloader_val, writer, "val", loss_val, epoch)

      # Checkpoints model weights.
      if epoch % save_model_frequency == 0:
        checkpointer.save(epoch)

      # Updates progress bar description.
      pbar_epoch.set_description(
          "TL: {:.2f} | VL: {:.2f} | THEORYMIN: {:.2f}".format(
              loss_train.detach().cpu().numpy().item(),
              loss_val.detach().cpu().numpy().item(),
              nll_limit,
          ))


if __name__ == "__main__":
  flags.mark_flag_as_required("dataset_dir")
  flags.mark_flag_as_required("output_dir")
  flags.mark_flag_as_required("num_epochs")
  app.run(main)



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines the interfaces for agents."""

import abc
import copy
import math
from collections import deque
from typing import Any
from typing import Mapping
from typing import Optional

import carla
import numpy as np
from absl import logging

import oatomobile
from oatomobile.simulators.carla import defaults
from oatomobile.util import carla as cutil


class SetPointAgent(oatomobile.Agent):
  """An agent that predicts setpoints and consumes them with a PID
  controller."""

  def __init__(
      self,
      environment: oatomobile.envs.CARLAEnv,
      *,
      setpoint_index: int = 5,
      replan_every_steps: int = 1,
      lateral_control_dict: Mapping[
          str, Any] = defaults.LATERAL_PID_CONTROLLER_CONFIG,
      longitudinal_control_dict: Mapping[
          str, Any] = defaults.LONGITUDINAL_PID_CONTROLLER_CONFIG,
      fixed_delta_seconds_between_setpoints: Optional[int] = None) -> None:
    """Constructs a setpoint-based agent.

    Args:
      environment: The navigation environment to spawn the agent.
      setpoint_index: The index of the point to cut-off the plan.
      replan_every_steps: The number of steps between subsequent call on the model.
      lateral_control_dict: The lateral PID controller's config.
      longitudinal_control_dict: The longitudinal PID controller's config.
      fixed_delta_seconds_between_setpoints: The time difference (in seconds)
        between the setpoints. It defaults to the fps of the simulator.
    """
    super(SetPointAgent, self).__init__(environment=environment)

    # References to the CARLA objects.
    self._vehicle = self._environment.simulator.hero
    self._world = self._vehicle.get_world()
    self._map = self._world.get_map()

    # Sets up PID controllers.
    dt = self._vehicle.get_world().get_settings().fixed_delta_seconds
    lateral_control_dict = lateral_control_dict.copy()
    lateral_control_dict.update({"dt": dt})
    logging.debug(
        "Lateral PID controller config {}".format(lateral_control_dict))
    longitudinal_control_dict = longitudinal_control_dict.copy()
    longitudinal_control_dict.update({"dt": dt})
    logging.debug("Longitudinal PID controller config {}".format(
        longitudinal_control_dict))
    self._vehicle_controller = oatomobile.VehiclePIDController(
        vehicle=self._vehicle,
        args_lateral=lateral_control_dict,
        args_longitudinal=longitudinal_control_dict,
    )

    # Sets agent's hyperparameters.
    self._setpoint_index = setpoint_index
    self._replan_every_steps = replan_every_steps
    self._fixed_delta_seconds_between_setpoints = fixed_delta_seconds_between_setpoints or dt

    # Inits agent's buffer of setpoints.
    self._setpoints_buffer = None
    self._steps_counter = 0

  @abc.abstractmethod
  def __call__(self, observation: oatomobile.Observations, *args,
               **kwargs) -> np.ndarray:
    """Returns the predicted plan in ego-coordinates, based on a model."""

  def act(self, observation: oatomobile.Observations, *args,
          **kwargs) -> oatomobile.Action:
    """Takes in an observation, samples from agent's policy, returns an
    action."""

    # Current measurements used for local2world2local transformations.
    current_location = observation["location"]
    current_rotation = observation["rotation"]

    if self._setpoints_buffer is None or self._steps_counter % self._replan_every_steps == 0:
      # Get agent predictions.
      predicted_plan_ego = self(copy.deepcopy(observation), *args,
                                **kwargs)  # [T, 3]
      # Transform plan to world coordinates
      predicted_plan_world = cutil.local2world(
          current_location=current_location,
          current_rotation=current_rotation,
          local_locations=predicted_plan_ego,
      )

      # Refreshes buffer.
      self._setpoints_buffer = predicted_plan_world

    else:
      # Pops first setpoint from the buffer.
      self._setpoints_buffer = self._setpoints_buffer[1:]

    # Registers setpoints for rendering.
    self._environment.unwrapped.simulator.sensor_suite.get(
        "predictions").predictions = cutil.world2local(
            current_location=current_location,
            current_rotation=current_rotation,
            world_locations=self._setpoints_buffer,
        )

    # Increments counter.
    self._steps_counter += 1

    # Calculates target speed by averaging speed in the `setpoint_index` window.
    target_speed = np.linalg.norm(
        np.diff(self._setpoints_buffer[:self._setpoint_index], axis=0),
        axis=1,
    ).mean() / (self._fixed_delta_seconds_between_setpoints)

    # Converts plan to PID controller setpoint.
    setpoint = self._map.get_waypoint(
        cutil.ndarray_to_location(self._setpoints_buffer[self._setpoint_index]))

    # Avoids getting stuck when spawned.
    if self._steps_counter <= 100:
      target_speed = 20.0 / 3.6

    # Run PID step.
    control = self._vehicle_controller.run_step(
        target_speed=target_speed *
        3.6,  # PID controller expects speed in km/h!
        waypoint=setpoint,
    )

    return control



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Public API for `oatomobile`."""

import os
import sys

from absl import logging

###############
# CARLA SETUP #
###############

# HACK(filangel): resolves https://github.com/carla-simulator/carla/issues/2132.
try:
  import torch
except ImportError:
  pass
try:
  import sonnet as snt
except ImportError:
  pass

# Enable CARLA PythonAPI to be accessed from `oatomobile`.
carla_path = os.getenv("CARLA_ROOT")
if carla_path is None:
  raise EnvironmentError(
      "Missing environment variable CARLA_ROOT, specify it before importing oatomobile"
  )

logging.debug("CARLA_ROOT={}".format(carla_path))
carla_python_api = os.path.join(
    carla_path,
    "PythonAPI",
    "carla",
)
if not os.path.exists(carla_python_api):
  raise ImportError("Missing CARLA installation at {}".format(carla_python_api))
sys.path.append(carla_python_api)

from agents.navigation.controller import VehiclePIDController  # pylint: disable=import-error
from agents.navigation.local_planner import \
    LocalPlanner  # pylint: disable=import-error
from agents.tools.misc import \
    compute_magnitude_angle  # pylint: disable=import-error
from agents.tools.misc import draw_waypoints  # pylint: disable=import-error
from agents.tools.misc import get_speed  # pylint: disable=import-error
from agents.tools.misc import \
    is_within_distance_ahead  # pylint: disable=import-error

###############

# HACK(filangel): matplotlib setup - remove before release.
import matplotlib
matplotlib.use("Agg")

# Benchmarks API.
from oatomobile.benchmarks.carnovel.benchmark import carnovel
# Core API.
from oatomobile.core.agent import Agent
from oatomobile.core.benchmark import Benchmark
from oatomobile.core.dataset import Dataset
from oatomobile.core.dataset import Episode
from oatomobile.core.dataset import tokens
from oatomobile.core.loop import EnvironmentLoop
from oatomobile.core.registry import registry
from oatomobile.core.rl import Env
from oatomobile.core.rl import FiniteHorizonWrapper
from oatomobile.core.rl import Metric
from oatomobile.core.rl import MonitorWrapper
from oatomobile.core.rl import ReturnsMetric
from oatomobile.core.rl import SaveToDiskWrapper
from oatomobile.core.rl import StepsMetric
from oatomobile.core.simulator import Action
from oatomobile.core.simulator import Observations
from oatomobile.core.simulator import Sensor
from oatomobile.core.simulator import SensorSuite
from oatomobile.core.simulator import SensorTypes
from oatomobile.core.simulator import Simulator

# Public API.
__all__ = (
    # CARLA Python API
    "compute_magnitude_angle",
    "draw_waypoints",
    "is_within_distance_ahead",
    "LocalPlanner",
    "VehiclePIDController",
    # OATomobile core API
    "Agent",
    "Benchmark",
    "Dataset",
    "EnvironmentLoop",
    "Episode",
    "tokens",
    "registry",
    "Env",
    "FiniteHorizonWrapper",
    "Metric",
    "MonitorWrapper",
    "ReturnsMetric",
    "StepsMetric",
    "SaveToDiskWrapper",
    "Action",
    "Observations",
    "Sensor",
    "SensorSuite",
    "Simulator",
)

#  ASCII art borrowed from dm-haiku.
#  __________________________________________
# / Please don't use these symbols they      \
# \ are not part of the OATomobile public API. /
#  ------------------------------------------
#         \   ^__^
#          \  (oo)\_______
#             (__)\       )\/\
#                 ||----w |
#                 ||     ||
#
del os
del sys
del logging
del carla_path
del carla_python_api
try:
  del snt
except NameError:
  pass
try:
  del torch
except NameError:
  pass



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Base implementation of agent inside `oatomobile`."""

import abc
from typing import Any

from oatomobile.core.rl import Env
from oatomobile.core.simulator import Action
from oatomobile.core.simulator import Observations


class Agent(abc.ABC):
  """An agent consists of an action-selection mechanism and an update rule."""

  def __init__(self, environment: Env, *args: Any, **kwargs: Any) -> None:
    """Constructs an agent."""
    self._environment = environment

  @abc.abstractmethod
  def act(
      self,
      observations: Observations,
  ) -> Action:
    """Samples an action from agent's policy, given observations."""

  def update(
      self,
      observations: Observations,
      action: Action,
      new_observations: Observations,
  ) -> None:
    """Updates the agent given a transition."""
    del observations
    del action
    del new_observations



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Default values for some parameters of the core API."""

import copy
import os

# CARLA simulator frames per second (FPS).
SIMULATOR_FPS = 20

# The front RGB camera configuration.
FRONT_CAMERA_RGB_SENSOR_CONFIG = {
    "attributes": {
        "image_size_x": "320",
        "image_size_y": "180",
        "fov": "90",
    },
    "actor": {
        "location": {
            "x": 0.0,
            "y": 0.0,
            "z": 2.3,
        },
        "rotation": {
            "pitch": 0,
        }
    },
}
# The rear RGB camera configuration.
REAR_CAMERA_RGB_SENSOR_CONFIG = {
    "attributes": {
        "image_size_x": "320",
        "image_size_y": "180",
        "fov": "90",
    },
    "actor": {
        "location": {
            "x": 0.0,
            "y": 0.0,
            "z": 2.3,
        },
        "rotation": {
            "pitch": 0,
            "yaw": 180,
        }
    },
}
# The left RGB camera configuration.
LEFT_CAMERA_RGB_SENSOR_CONFIG = {
    "attributes": {
        "image_size_x": "320",
        "image_size_y": "180",
        "fov": "90",
    },
    "actor": {
        "location": {
            "x": 0.0,
            "y": 0.0,
            "z": 2.3,
        },
        "rotation": {
            "pitch": 0,
            "yaw": 270,
        }
    },
}
# The right RGB camera configuration.
RIGHT_CAMERA_RGB_SENSOR_CONFIG = {
    "attributes": {
        "image_size_x": "320",
        "image_size_y": "180",
        "fov": "90",
    },
    "actor": {
        "location": {
            "x": 0.0,
            "y": 0.0,
            "z": 2.3,
        },
        "rotation": {
            "pitch": 0,
            "yaw": 90,
        }
    },
}

# The bird-view RGB/CityScapes camera configuration.
BIRD_VIEW_CAMERA_RGB_SENSOR_CONFIG = {
    "attributes": {
        "image_size_x": "200",
        "image_size_y": "200",
        "fov": "90",
    },
    "actor": {
        "location": {
            "x": 0.0,
            "y": 0.0,
            "z": 25.0,
        },
        "rotation": {
            "pitch": 270,
        }
    },
}
BIRD_VIEW_CAMERA_CITYSCAPES_SENSOR_CONFIG = copy.deepcopy(
    BIRD_VIEW_CAMERA_RGB_SENSOR_CONFIG)

# The LIDAR configuration.
LIDAR_SENSOR_CONFIG = {
    "attributes": {
        "range": "5000",
        "points_per_second": str(SIMULATOR_FPS * 10000),
        "rotation_frequency": str(SIMULATOR_FPS),
        "upper_fov": "10",
        "lower_fov": "-30",
    },
    "actor": {
        "location": {
            "x": 0.0,
            "y": 0.0,
            "z": 2.5,
        },
    },
}

# The goal sensor configuration.
GOAL_SENSOR_CONFIG = {
    "num_goals": 10,
    "sampling_radius": 2.0,
    "replan_every_steps": 5,
}

# The game state configuration.
GAME_STATE_CONFIG = {
    "margin": 150,
    "scale": 1.0,
    "pixels_per_meter": 5,
}

# Default sensors.
CARLA_SENSORS = (
    "goal",
    "front_camera_rgb",
    "lidar",
    "bird_view_camera_cityscapes",
    "bird_view_camera_rgb",
    "control",
    "location",
    "rotation",
    "velocity",
    "collision",
    "lane_invasion",
    "is_at_traffic_light",
    "traffic_light_state",
)

# The time interval before stopping the search for the carla server.
CARLA_CLIENT_TIMEOUT = 20.0

# The time interval before stopping the search to the queue.
QUEUE_TIMEOUT = 2.0

# Available CARLA towns.
AVAILABLE_CARLA_TOWNS = (
    "Town01",
    "Town02",
    "Town03",
    "Town04",
    "Town05",
)

# Speed configuration of autopilot.
TARGET_SPEED = 20.0

# PID controllers configuration.
LATERAL_PID_CONTROLLER_CONFIG = {
    'K_P': 1.95,
    'K_D': 0.01,
    'K_I': 1.4,
    'dt': 1.0 / SIMULATOR_FPS,
}
LONGITUDINAL_PID_CONTROLLER_CONFIG = {
    'K_P': 1.0,
    'K_D': 0,
    'K_I': 1.0,
    'dt': 1.0 / SIMULATOR_FPS,
}

# The number of simulator steps before termination.
MAX_EPISODE_STEPS = int(1e4)



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines a simple PID-controller-based autopilot."""

import math
import random
from typing import Any
from typing import Optional
from typing import Tuple

import carla
from absl import logging

import oatomobile
from oatomobile.simulators.carla import defaults
from oatomobile.util import carla as cutil


class AutopilotAgent(oatomobile.Agent):
  """An autopilot agent, based on the official implementation of
  `carla.PythonAPI.agents.navigation.basic_agent.BasicAgent`"""

  def __init__(self,
               environment: oatomobile.envs.CARLAEnv,
               *,
               proximity_tlight_threshold: float = 5.0,
               proximity_vehicle_threshold: float = 10.0,
               noise: float = 0.1) -> None:
    """Constructs an autopilot agent.

    Args:
      environment: The navigation environment to spawn the agent.
      proximity_tlight_threshold: The threshold (in metres) to
        the traffic light.
      proximity_vehicle_threshold: The threshold (in metres) to
        the front vehicle.
      noise: The percentage of random actions.
    """
    super(AutopilotAgent, self).__init__(environment=environment)

    # References to the CARLA objects.
    self._vehicle = self._environment.simulator.hero
    self._world = self._vehicle.get_world()
    self._map = self._world.get_map()

    # Agent hyperparametres.
    self._proximity_tlight_threshold = proximity_tlight_threshold
    self._proximity_vehicle_threshold = proximity_vehicle_threshold
    self._hop_resolution = 2.0
    self._path_seperation_hop = 2
    self._path_seperation_threshold = 0.5
    self._target_speed = defaults.TARGET_SPEED
    self._noise = noise

    # The internal state of the agent.
    self._last_traffic_light = None

    # Local planner, including the PID controllers.
    dt = self._vehicle.get_world().get_settings().fixed_delta_seconds
    lateral_control_dict = defaults.LATERAL_PID_CONTROLLER_CONFIG.copy()
    lateral_control_dict.update({"dt": dt})
    # TODO(filangel): tune the parameters for FPS != 20
    self._local_planner = oatomobile.LocalPlanner(
        self._vehicle,
        opt_dict=dict(
            target_speed=self._target_speed,
            dt=dt,
        ),
    )

    # Set agent's dsestination.
    if hasattr(self._environment.unwrapped.simulator, "destination"):
      self._set_destination(
          self._environment.unwrapped.simulator.destination.location)

  def act(
      self,
      observation: oatomobile.Observations,
  ) -> oatomobile.Action:
    """Takes in an observation, samples from agent's policy, returns an
    action."""
    # Remove unused arguments.
    del observation

    # Random action branch.
    if random.random() < self._noise:
      return carla.VehicleControl(  # pylint: disable=no-member
          **{
              k: float(v)
              for (k, v) in self._environment.action_space.sample().items()
          })
    # Normal autopilot action.
    else:
      return self._run_step()

  def _run_step(
      self,
      debug: bool = False,
  ) -> carla.VehicleControl:  # pylint: disable=no-member
    """Executes one step of navigation."""

    # is there an obstacle in front of us?
    hazard_detected = False

    # retrieve relevant elements for safe navigation, i.e.: traffic lights
    # and other vehicles
    actor_list = self._world.get_actors()
    vehicle_list = actor_list.filter("*vehicle*")
    lights_list = actor_list.filter("*traffic_light*")

    # check possible obstacles
    vehicle_state, vehicle = self._is_vehicle_hazard(vehicle_list)
    if vehicle_state:
      if debug:
        logging.debug('!!! VEHICLE BLOCKING AHEAD [{}])'.format(vehicle.id))

      hazard_detected = True

    # check for the state of the traffic lights
    light_state, traffic_light = self._is_light_red(lights_list)
    if light_state:
      if debug:
        logging.debug('=== RED LIGHT AHEAD [{}])'.format(traffic_light.id))

      hazard_detected = True

    if hazard_detected:
      control = carla.VehicleControl()  # pylint: disable=no-member
      control.steer = 0.0
      control.throttle = 0.0
      control.brake = 1.0
      control.hand_brake = False
    else:
      # standard local planner behavior
      control = self._local_planner.run_step(debug=debug)

    return control

  def _set_destination(
      self,
      destination: carla.Location,  # pylint: disable=no-member
  ) -> None:
    """Generates the global plan for the agent.

    Args:
      destination: The location of the new destination.
    """
    # Set vehicle's current location as start for the plan.
    origin = self._vehicle.get_location()
    start_waypoint = self._map.get_waypoint(origin).transform.location
    end_waypoint = self._map.get_waypoint(destination).transform.location
    # Calculate the plan.
    waypoints, roadoptions, _ = cutil.global_plan(
        world=self._world,
        origin=start_waypoint,
        destination=end_waypoint,
    )
    # Mutate the local planner's global plan.
    self._local_planner.set_global_plan(list(zip(waypoints, roadoptions)))

  def _is_vehicle_hazard(
      self,
      vehicle_list,
  ) -> Tuple[bool, Optional[carla.Vehicle]]:  # pylint: disable=no-member
    """It detects if a vehicle in the scene can be dangerous for the ego
    vehicle's current plan.

    Args:
      vehicle_list: List of potential vehicles (obstancles) to check.

    Returns:
      vehicle_ahead: If True a vehicle ahead blocking us and False otherwise.
      vehicle: The blocker vehicle itself.
    """

    ego_vehicle_location = self._vehicle.get_location()
    ego_vehicle_waypoint = self._map.get_waypoint(ego_vehicle_location)

    for target_vehicle in vehicle_list:
      # do not account for the ego vehicle.
      if target_vehicle.id == self._vehicle.id:
        continue

      # if the object is not in our lane it's not an obstacle.
      target_vehicle_waypoint = self._map.get_waypoint(
          target_vehicle.get_location())
      if target_vehicle_waypoint.road_id != ego_vehicle_waypoint.road_id or \
              target_vehicle_waypoint.lane_id != ego_vehicle_waypoint.lane_id:
        continue

      loc = target_vehicle.get_location()
      if oatomobile.is_within_distance_ahead(
          loc,
          ego_vehicle_location,
          self._vehicle.get_transform().rotation.yaw,
          self._proximity_vehicle_threshold,
      ):
        return (True, target_vehicle)

    return (False, None)

  def _is_light_red(
      self,
      lights_list,
  ) -> Tuple[bool, Any]:  # pylint: disable=no-member
    """It detects if the light in the scene is red.

    Args:
      lights_list: The list containing TrafficLight objects.

    Returns:
      light_ahead: If True a traffic light ahead is read and False otherwise.
      traffic_light: The traffic light object ahead itself.
    """
    if self._map.name == 'Town01' or self._map.name == 'Town02':
      return self._is_light_red_europe_style(lights_list)
    else:
      return self._is_light_red_us_style(lights_list)

  def _is_light_red_europe_style(self, lights_list):
    """This method is specialized to check European style traffic lights."""
    ego_vehicle_location = self._vehicle.get_location()
    ego_vehicle_waypoint = self._map.get_waypoint(ego_vehicle_location)

    for traffic_light in lights_list:
      object_waypoint = self._map.get_waypoint(traffic_light.get_location())
      if object_waypoint.road_id != ego_vehicle_waypoint.road_id or \
              object_waypoint.lane_id != ego_vehicle_waypoint.lane_id:
        continue

      loc = traffic_light.get_location()
      if oatomobile.is_within_distance_ahead(
          loc,
          ego_vehicle_location,
          self._vehicle.get_transform().rotation.yaw,
          self._proximity_tlight_threshold,
      ):
        if traffic_light.state == carla.TrafficLightState.Red:  # pylint: disable=no-member
          return (True, traffic_light)

    return (False, None)

  def _is_light_red_us_style(self, lights_list, debug=False):
    """This method is specialized to check US style traffic lights."""
    ego_vehicle_location = self._vehicle.get_location()
    ego_vehicle_waypoint = self._map.get_waypoint(ego_vehicle_location)

    if ego_vehicle_waypoint.is_junction:
      # It is too late. Do not block the intersection! Keep going!
      return (False, None)

    if self._local_planner.target_waypoint is not None:
      if self._local_planner.target_waypoint.is_junction:
        min_angle = 180.0
        sel_magnitude = 0.0
        sel_traffic_light = None
        for traffic_light in lights_list:
          loc = traffic_light.get_location()
          magnitude, angle = oatomobile.compute_magnitude_angle(
              loc, ego_vehicle_location,
              self._vehicle.get_transform().rotation.yaw)
          if magnitude < 60.0 and angle < min(25.0, min_angle):
            sel_magnitude = magnitude
            sel_traffic_light = traffic_light
            min_angle = angle

        if sel_traffic_light is not None:
          if debug:
            logging.debug('=== Magnitude = {} | Angle = {} | ID = {}'.format(
                sel_magnitude, min_angle, sel_traffic_light.id))

          if self._last_traffic_light is None:
            self._last_traffic_light = sel_traffic_light

          if self._last_traffic_light.state == carla.TrafficLightState.Red:  # pylint: disable=no-member
            return (True, self._last_traffic_light)
        else:
          self._last_traffic_light = None

    return (False, None)

  def _get_trafficlight_trigger_location(
      self,
      traffic_light,
  ) -> carla.Location:  # pylint: disable=no-member
    """Calculates the yaw of the waypoint that represents the trigger volume of
    the traffic light."""

    def rotate_point(point, radians):
      """Rotates a given point by a given angle."""
      rotated_x = math.cos(radians) * point.x - math.sin(radians) * point.y
      rotated_y = math.sin(radians) * point.x - math.cos(radians) * point.y

      return carla.Vector3D(rotated_x, rotated_y, point.z)  # pylint: disable=no-member

    base_transform = traffic_light.get_transform()
    base_rot = base_transform.rotation.yaw
    area_loc = base_transform.transform(traffic_light.trigger_volume.location)
    area_ext = traffic_light.trigger_volume.extent

    point = rotate_point(
        carla.Vector3D(0, 0, area_ext.z),  # pylint: disable=no-member
        math.radians(base_rot),
    )
    point_location = area_loc + carla.Location(x=point.x, y=point.y)  # pylint: disable=no-member

    return carla.Location(point_location.x, point_location.y, point_location.z)  # pylint: disable=no-member



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines a discriminative model for the conditional imitation learner."""

from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import torch
import torch.distributions as D
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from oatomobile.baselines.torch import transforms
from oatomobile.baselines.torch.models import MLP
from oatomobile.baselines.torch.models import MobileNetV2
from oatomobile.baselines.torch.typing import ArrayLike
from oatomobile.core.typing import ShapeLike


class ImitativeModel(nn.Module):
  """A `PyTorch` implementation of an imitative model."""

  def __init__(
      self,
      output_shape: ShapeLike = (4, 2),
  ) -> None:
    """Constructs a simple behavioural cloning model.

    Args:
      output_shape: The shape of the base and
        data distribution (a.k.a. event_shape).
    """
    super(ImitativeModel, self).__init__()
    self._output_shape = output_shape

    # Initialises the base distribution.
    self._base_dist = D.MultivariateNormal(
        loc=torch.zeros(self._output_shape[-2] * self._output_shape[-1]),  # pylint: disable=no-member
        scale_tril=torch.eye(self._output_shape[-2] * self._output_shape[-1]),  # pylint: disable=no-member
    )

    # The convolutional encoder model.
    self._encoder = MobileNetV2(num_classes=128, in_channels=2)

    # Merges the encoded features and the vector inputs.
    self._merger = MLP(
        input_size=128 + 3 + 1 + 1,
        output_sizes=[64, 64, 64],
        activation_fn=nn.ReLU,
        dropout_rate=None,
        activate_final=True,
    )

    # The decoder recurrent network used for the sequence generation.
    self._decoder = nn.GRUCell(input_size=2, hidden_size=64)

    # The output head.
    self._locscale = MLP(
        input_size=64,
        output_sizes=[32, 4],
        activation_fn=nn.ReLU,
        dropout_rate=None,
        activate_final=False,
    )

  def to(self, *args, **kwargs):
    """Handles non-parameter tensors when moved to a new device."""
    self = super().to(*args, **kwargs)
    self._base_dist = D.MultivariateNormal(
        loc=self._base_dist.mean.to(*args, **kwargs),
        scale_tril=self._base_dist.scale_tril.to(*args, **kwargs),
    )
    return self

  def forward(
      self,
      num_steps: int,
      goal: Optional[torch.Tensor] = None,
      lr: float = 1e-1,
      epsilon: float = 1.0,
      **context: torch.Tensor) -> Union[torch.Tensor, Sequence[torch.Tensor]]:
    """Returns a local mode from the posterior.

    Args:
      num_steps: The number of gradient-descent steps for finding the mode.
      goal: The locations of the the goals.
      epsilon: The tolerance parameter for the goal.
      context: (keyword arguments) The conditioning
        variables used for the conditional flow.

    Returns:
      A mode from the posterior, with shape `[D, 2]`.
    """
    if not "visual_features" in context:
      raise ValueError("Missing `visual_features` keyword argument.")
    batch_size = context["visual_features"].shape[0]

    # Sets initial sample to base distribution's mean.
    x = self._base_dist.sample().clone().detach().repeat(batch_size, 1).view(
        batch_size,
        *self._output_shape,
    )
    x.requires_grad = True

    # The contextual parameters, caches for efficiency.
    z = self._params(**context)

    # Initialises a gradient-based optimiser.
    optimizer = optim.Adam(params=[x], lr=lr)

    # Stores the best values.
    x_best = x.clone()
    loss_best = torch.ones(()).to(x.device) * 1000.0  # pylint: disable=no-member

    for _ in range(num_steps):
      # Resets optimizer's gradients.
      optimizer.zero_grad()
      # Operate on `y`-space.
      y, _ = self._forward(x=x, z=z)
      # Calculates imitation prior.
      _, log_prob, logabsdet = self._inverse(y=y, z=z)
      imitation_prior = torch.mean(log_prob - logabsdet)  # pylint: disable=no-member
      # Calculates goal likelihodd.
      goal_likelihood = 0.0
      if goal is not None:
        goal_likelihood = self._goal_likelihood(y=y, goal=goal, epsilon=epsilon)
      loss = -(imitation_prior + goal_likelihood)
      # Backward pass.
      loss.backward(retain_graph=True)
      # Performs a gradient descent step.
      optimizer.step()
      # Book-keeping
      if loss < loss_best:
        x_best = x.clone()
        loss_best = loss.clone()

    y, _ = self._forward(x=x_best, z=z)

    return y

  def _goal_likelihood(self, y: torch.Tensor, goal: torch.Tensor,
                       **hyperparams) -> torch.Tensor:
    """Returns the goal-likelihood of a plan `y`, given `goal`.

    Args:
      y: A plan under evaluation, with shape `[B, T, 2]`.
      goal: The goal locations, with shape `[B, K, 2]`.
      hyperparams: (keyword arguments) The goal-likelihood hyperparameters.

    Returns:
      The log-likelihodd of the plan `y` under the `goal` distribution.
    """
    # Parses tensor dimensions.
    B, K, _ = goal.shape

    # Fetches goal-likelihood hyperparameters.
    epsilon = hyperparams.get("epsilon", 1.0)

    # TODO(filangel): implement other goal likelihoods from the DIM paper
    # Initializes the goal distribution.
    goal_distribution = D.MixtureSameFamily(
        mixture_distribution=D.Categorical(
            probs=torch.ones((B, K)).to(goal.device)),  # pylint: disable=no-member
        component_distribution=D.Independent(
            D.Normal(loc=goal, scale=torch.ones_like(goal) * epsilon),  # pylint: disable=no-member
            reinterpreted_batch_ndims=1,
        ))

    return torch.mean(goal_distribution.log_prob(y[:, -1, :]), dim=0)  # pylint: disable=no-member

  def _params(self, **context: torch.Tensor) -> torch.Tensor:
    """Returns the contextual parameters of the conditional density estimator.

    Args:
      visual_features: The visual input, with shape `[B, H, W, 3]`.
      velocity: The vehicle's current velocity, with shape `[B, 3]`.
      is_at_traffic_light: The flag for being close to a traffic light, with
        shape `[B, 1]`.
      traffic_light_state: The state of the nearest traffic light, with shape
        `[B, 1]`.

    Returns:
      The contextual parameters of the conditional density estimator.
    """

    # Parses context variables.
    if not "visual_features" in context:
      raise ValueError("Missing `visual_features` keyword argument.")
    if not "velocity" in context:
      raise ValueError("Missing `velocity` keyword argument.")
    if not "is_at_traffic_light" in context:
      raise ValueError("Missing `is_at_traffic_light` keyword argument.")
    if not "traffic_light_state" in context:
      raise ValueError("Missing `traffic_light_state` keyword argument.")
    visual_features = context.get("visual_features")
    velocity = context.get("velocity")
    is_at_traffic_light = context.get("is_at_traffic_light")
    traffic_light_state = context.get("traffic_light_state")

    # Encodes the visual input.
    visual_features = self._encoder(visual_features)

    # Merges visual input logits and vector inputs.
    visual_features = torch.cat(  # pylint: disable=no-member
        tensors=[
            visual_features,
            velocity,
            is_at_traffic_light,
            traffic_light_state,
        ],
        dim=-1,
    )

    # The decoders initial state.
    visual_features = self._merger(visual_features)

    return visual_features

  def _forward(
      self,
      x: torch.Tensor,
      z: torch.Tensor,
  ) -> Tuple[torch.Tensor, torch.Tensor]:
    """Transforms samples from the base distribution to the data distribution.

    Args:
      x: Samples from the base distribution, with shape `[B, D]`.
      z: The contextual parameters of the conditional density estimator, with
        shape `[B, K]`.

    Returns:
      y: The sampels from the push-forward distribution,
        with shape `[B, D]`.
      logabsdet: The log absolute determinant of the Jacobian,
        with shape `[B]`.
    """

    # Output containers.
    y = list()
    scales = list()

    # Initial input variable.
    y_tm1 = torch.zeros(  # pylint: disable=no-member
        size=(z.shape[0], self._output_shape[-1]),
        dtype=z.dtype,
    ).to(z.device)

    for t in range(x.shape[-2]):
      x_t = x[:, t, :]

      # Unrolls the GRU.
      z = self._decoder(y_tm1, z)

      # Predicts the location and scale of the MVN distribution.
      dloc_scale = self._locscale(z)
      dloc = dloc_scale[..., :2]
      scale = F.softplus(dloc_scale[..., 2:]) + 1e-3

      # Data distribution corresponding sample.
      y_t = (y_tm1 + dloc) + scale * x_t

      # Update containers.
      y.append(y_t)
      scales.append(scale)
      y_tm1 = y_t

    # Prepare tensors, reshape to [B, T, 2].
    y = torch.stack(y, dim=-2)  # pylint: disable=no-member
    scales = torch.stack(scales, dim=-2)  # pylint: disable=no-member

    # Log absolute determinant of Jacobian.
    logabsdet = torch.log(torch.abs(torch.prod(scales, dim=-2)))  # pylint: disable=no-member
    logabsdet = torch.sum(logabsdet, dim=-1)  # pylint: disable=no-member

    return y, logabsdet

  def _inverse(
      self,
      y: torch.Tensor,
      z: torch.Tensor,
  ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """Transforms samples from the data distribution to the base distribution.

    Args:
      y: Samples from the data distribution, with shape `[B, D]`.
      z: The contextual parameters of the conditional density estimator, with shape
        `[B, K]`.

    Returns:
      x: The sampels from the base distribution,
        with shape `[B, D]`.
      log_prob: The log-likelihood of the samples under
        the base distibution probability, with shape `[B]`.
      logabsdet: The log absolute determinant of the Jacobian,
        with shape `[B]`.
    """

    # Output containers.
    x = list()
    scales = list()

    # Initial input variable.
    y_tm1 = torch.zeros(  # pylint: disable=no-member
        size=(z.shape[0], self._output_shape[-1]),
        dtype=z.dtype,
    ).to(z.device)

    for t in range(y.shape[-2]):
      y_t = y[:, t, :]

      # Unrolls the GRU.
      z = self._decoder(y_tm1, z)

      # Predicts the location and scale of the MVN distribution.
      dloc_scale = self._locscale(z)
      dloc = dloc_scale[..., :2]
      scale = F.softplus(dloc_scale[..., 2:]) + 1e-3

      # Base distribution corresponding sample.
      x_t = (y_t - (y_tm1 + dloc)) / scale

      # Update containers.
      x.append(x_t)
      scales.append(scale)
      y_tm1 = y_t

    # Prepare tensors, reshape to [B, T, 2].
    x = torch.stack(x, dim=-2)  # pylint: disable=no-member
    scales = torch.stack(scales, dim=-2)  # pylint: disable=no-member

    # Log likelihood under base distribution.
    log_prob = self._base_dist.log_prob(x.view(x.shape[0], -1))

    # Log absolute determinant of Jacobian.
    logabsdet = torch.log(  # pylint: disable=no-member
        torch.abs(torch.prod(  # pylint: disable=no-member
            scales, dim=-1)))  # determinant == product over xy-coordinates
    logabsdet = torch.sum(logabsdet, dim=-1)  # sum over T dimension # pylint: disable=no-member

    return x, log_prob, logabsdet

  def transform(
      self,
      sample: Mapping[str, ArrayLike],
  ) -> Mapping[str, torch.Tensor]:
    """Prepares variables for the interface of the model.

    Args:
      sample: (keyword arguments) The raw sample variables.

    Returns:
      The processed sample.
    """

    # Preprocesses the target variables.
    if "player_future" in sample:
      sample["player_future"] = transforms.downsample_target(
          player_future=sample["player_future"],
          num_timesteps_to_keep=self._output_shape[-2],
      )

    # Renames `lidar` to `visual_features`.
    if "lidar" in sample:
      sample["visual_features"] = sample.pop("lidar")

    # Preprocesses the visual features.
    if "visual_features" in sample:
      sample["visual_features"] = transforms.transpose_visual_features(
          transforms.downsample_visual_features(
              visual_features=sample["visual_features"],
              output_shape=(100, 100),
          ))

    return sample



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
from importlib import util as import_util

from setuptools import find_packages
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README.mf file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
  long_description = f.read()

# Get the version from metadata.
spec = import_util.spec_from_file_location(
    "_metadata",
    "oatomobile/_metadata.py",
)
_metadata = import_util.module_from_spec(spec)
spec.loader.exec_module(_metadata)
version = _metadata.__version__

setup(
    name="oatomobile",
    version=version,
    description=
    "The oatomobile is a tool for developing and testing driving agents on the CARLA simulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/oatml/oatomobile",
    author="Oxford Applied and Theoretical Machine Learning Group",
    author_email="oatml@googlegroups.com",
    license="Apache License, Version 2.0",
    packages=find_packages(),
    install_requires=[
        "absl-py==0.9.0",
        "gym==0.17.1",
        "pygame==1.9.6",
        "matplotlib==3.0.3",
        "seaborn==0.9.1",
        "pandas==0.25.3",
        "scipy==1.4.1",
        "transforms3d==0.3.1",
        "tqdm==4.42.1",
        "wget==3.2",
        "networkx==2.4",
        "imageio==2.8.0",
        "tabulate==0.8.7",
        "scikit-image==0.15.0",
    ],
    tests_require=[
        "pytest",
    ],
    # See oatomobile/baselines/README.md for more information.
    extras_require={
        # Additional requirements for TensorFlow baselines.
        "tf": [
            "dm-sonnet==2.0.0",
            "tensorflow==2.2.0",
            "tensorflow_probability==0.9.0",
            "tensorboard==2.2.0",
        ],
        # Additional requirements for PyTorch baselines.
        "torch": [
            "torch==1.5.1",
            "torchvision==0.6.0",
            "tensorboard==2.2.0",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Runs the blind rule-based agent on CARLA."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import os

from absl import app
from absl import flags
from absl import logging

import oatomobile
import oatomobile.envs
from oatomobile.baselines.rulebased.blind.agent import BlindAgent

logging.set_verbosity(logging.DEBUG)
FLAGS = flags.FLAGS
flags.DEFINE_enum(
    name="town",
    default="Town01",
    enum_values=[
        "Town01",
        "Town02",
        "Town03",
        "Town04",
        "Town05",
    ],
    help="The CARLA town id.",
)
flags.DEFINE_list(
    name="sensors",
    default=[
        "goal",
        "velocity",
        "bird_view_camera_cityscapes",
        "bird_view_camera_rgb",
        "front_camera_rgb",
        "lidar",
    ],
    help="The list of recorded sensors.",
)
flags.DEFINE_integer(
    name="max_episode_steps",
    default=None,
    help="The number of steps in the simulator.",
)
flags.DEFINE_string(
    name="output_dir",
    default=None,
    help="The full path to the output directory.",
)
flags.DEFINE_bool(
    name="render",
    default=False,
    help="If True it spawn the `PyGame` display.",
)


def main(argv):
  # Debugging purposes.
  logging.debug(argv)
  logging.debug(FLAGS)

  # Parses command line arguments.
  town = FLAGS.town
  sensors = FLAGS.sensors
  max_episode_steps = FLAGS.max_episode_steps
  output_dir = FLAGS.output_dir
  render = FLAGS.render

  try:
    # Setups the environment.
    env = oatomobile.envs.CARLANavEnv(
        town=town,
        origin=0,
        destination=25,
        fps=20,
        sensors=sensors,
    )
    if max_episode_steps is not None:
      env = oatomobile.FiniteHorizonWrapper(
          env,
          max_episode_steps=max_episode_steps,
      )
    if output_dir is not None:
      env = oatomobile.SaveToDiskWrapper(env, output_dir=output_dir)

    # Runs the environment loop.
    oatomobile.EnvironmentLoop(
        agent_fn=functools.partial(BlindAgent, setpoint_index=2),
        environment=env,
        render_mode="human" if render else "none",
    ).run()

  except Exception as msg:
    logging.error(msg)

  finally:
    # Garbage collector.
    try:
      env.close()
    except NameError:
      pass


if __name__ == "__main__":
  app.run(main)



# Copyright 2020 The OATomobile Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Module and layer definitions used across the `PyTorch` models."""

from typing import Callable
from typing import Optional
from typing import Sequence

import torch
import torch.nn as nn


class MobileNetV2(nn.Module):
  """A `PyTorch Hub` MobileNetV2 model wrapper."""

  def __init__(
      self,
      num_classes: int,
      in_channels: int = 3,
  ) -> None:
    """Constructs a MobileNetV2 model."""
    super(MobileNetV2, self).__init__()

    self._model = torch.hub.load(
        github="pytorch/vision:v0.6.0",
        model="mobilenet_v2",
        num_classes=num_classes,
    )

    # HACK(filangel): enables non-RGB visual features.
    _tmp = self._model.features._modules['0']._modules['0']
    self._model.features._modules['0']._modules['0'] = nn.Conv2d(
        in_channels=in_channels,
        out_channels=_tmp.out_channels,
        kernel_size=_tmp.kernel_size,
        stride=_tmp.stride,
        padding=_tmp.padding,
        bias=_tmp.bias,
    )

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """Forward pass from the MobileNetV2."""
    return self._model(x)


class MLP(nn.Module):
  """A simple multi-layer perceptron module."""

  def __init__(
      self,
      input_size: int,
      output_sizes: Sequence[int],
      activation_fn: Callable[[], nn.Module] = nn.ReLU,
      dropout_rate: Optional[float] = None,
      activate_final: bool = False,
  ) -> None:
    """Constructs a simple multi-layer-perceptron.

    Args:
      input_size: The size of the input features.
      output_sizes: Sequence of layer sizes.
      w_init: Initializer for Linear weights.
      activation: Activation function to apply between linear layers. Defaults
        to ReLU.
      dropout_rate: Dropout rate to apply, a rate of `None` (the default) or `0`
        means no dropout will be applied.
      activate_final: Whether or not to activate the final layer of the MLP.
    """
    super(MLP, self).__init__()

    layers = list()
    for in_features, out_features in zip(
        [input_size] + list(output_sizes)[:-2],
        output_sizes[:-1],
    ):
      # Fully connected layer.
      layers.append(nn.Linear(in_features, out_features))
      # Activation layer.
      layers.append(activation_fn(inplace=True))
      # (Optional) dropout layer.
      if dropout_rate is not None:
        layers.append(nn.Dropout(p=dropout_rate, inplace=True))
    # Final layer.
    layers.append(nn.Linear(output_sizes[-2], output_sizes[-1]))
    # (Optional) output activation layer.
    if activate_final:
      layers.append(activation_fn(inplace=True))

    self._model = nn.Sequential(*layers)

  def forward(self, x: torch.Tensor) -> torch.Tensor:
    """Forward pass from the MLP."""
    return self._model(x)
